"""
BLOOM LIMS Domain - Base Classes

This module contains the BloomObj base class which provides core functionality
for all BLOOM domain objects.

Extracted from bloom_lims/bobjs.py for better code organization.
"""

import os
import json
import re
import random
import string
import logging
import subprocess
import socket
from datetime import datetime, timedelta, date, UTC
from pathlib import Path
import urllib.parse

import pytz
import yaml
import boto3
import requests
from botocore.exceptions import NoCredentialsError, PartialCredentialsError

from sqlalchemy import (
    and_,
    desc,
    text,
    or_,
    cast,
    func,
    DateTime,
    select,
)
from sqlalchemy.orm.attributes import flag_modified

import zebra_day.print_mgr as zdpm

from bloom_lims.domain.utils import (
    get_datetime_string,
    generate_random_string,
    update_recursive as _update_recursive,
    unique_non_empty_strings,
)

# Try to import carrier tracking module (replaced fedex_tracking_day with daylily_carrier_tracking)
try:
    import daylily_carrier_tracking as FTD
except Exception:
    FTD = None

# Universal printer behavior
PGLOBAL = False if os.environ.get("PGLOBAL", False) else True

logger = logging.getLogger(__name__)


class BloomObj:
    def __init__(
        self, bdb, is_deleted=False, cfg_printers=False, cfg_fedex=False
    ):  # ERROR -- the is_deleted flag should be set, I think, at the db level...
        self.logger = logging.getLogger(__name__ + ".BloomObj")
        self.logger.debug("Instantiating BloomObj")

        # Zebra Day Print Manager
        self.zpld = zdpm.zpl()
        if cfg_printers:
            self._config_printers()


        # Move the  fedex and zebra stuff outside these objs
        self.track_fedex = None
        if cfg_fedex:
            fedex_key = os.environ.get("FEDEX_API_KEY")
            fedex_secret = os.environ.get("FEDEX_SECRET")
            if fedex_key and fedex_secret and FTD is not None:
                try:
                    self.track_fedex = FTD.FedexTracker()
                except Exception as e:
                    self.logger.warning(
                        "FedEx tracking disabled; failed to initialize client: %s", e
                    )
            else:
                self.logger.info(
                    "FedEx tracking disabled; missing FEDEX_API_KEY/FEDEX_SECRET"
                )
    
        self._bdb = bdb
        self.is_deleted = is_deleted
        self.session = bdb.session
        self.Base = bdb.Base
        app_username = str(getattr(bdb, "app_username", "") or "").strip()
        self._actor_user_id = None
        self._actor_email = app_username if "@" in app_username else None

    def set_actor_context(self, *, user_id=None, email=None):
        """Set optional actor identity context used for interaction lineage tracking."""
        user_id_str = str(user_id or "").strip()
        email_str = str(email or "").strip()
        self._actor_user_id = user_id_str or None
        self._actor_email = email_str or self._actor_email
        return self

    def _extract_actor_from_curr_user(self, curr_user):
        actor_user_id = None
        actor_email = None
        if isinstance(curr_user, dict):
            actor_user_id = (
                curr_user.get("cognito_sub")
                or curr_user.get("sub")
                or curr_user.get("user_id")
            )
            actor_email = curr_user.get("email")
        elif isinstance(curr_user, str):
            curr = curr_user.strip()
            if "@" in curr:
                actor_email = curr
            elif curr:
                actor_user_id = curr
        return (str(actor_user_id).strip() or None, str(actor_email).strip() or None)

    def _instance_properties(self, instance):
        payload = instance.json_addl or {}
        if isinstance(payload, str):
            try:
                payload = json.loads(payload)
            except Exception:
                payload = {}
        if not isinstance(payload, dict):
            payload = {}
        props = payload.get("properties")
        if not isinstance(props, dict):
            props = {}
            payload["properties"] = props
            instance.json_addl = payload
        return payload, props

    def _ensure_user_actor_template(self):
        templates = self.query_template_by_component_v2(
            "actor", "user", "bloom-user", "1.0"
        )
        if templates:
            return templates[0]

        template = self.Base.classes.generic_template(
            name="Bloom User",
            polymorphic_discriminator="generic_template",
            category="actor",
            type="user",
            subtype="bloom-user",
            version="1.0",
            instance_prefix="AX",
            instance_polymorphic_identity="generic_instance",
            json_addl={
                "description": "Bloom authenticated user projection",
                "properties": {
                    "name": "Bloom User",
                    "email": "",
                    "cognito_sub": "",
                    "comments": "",
                },
                "expected_inputs": [],
                "singleton": "0",
                "expected_outputs": [],
                "action_groups": {},
                "action_imports": {
                    "core": {
                        "group_order": "1",
                        "group_name": "Core Actions",
                        "actions": {"action/core/*/1.0/": {}},
                    }
                },
                "instantiation_layouts": [],
            },
            bstatus="active",
            is_singleton=False,
            is_deleted=False,
        )
        self.session.add(template)
        self.session.flush()
        return template

    def _upsert_user_actor(self, *, user_id=None, email=None):
        resolved_user_id = str(user_id or "").strip() or None
        resolved_email = str(email or "").strip() or None
        if not resolved_user_id and not resolved_email:
            return None

        actors = (
            self.session.query(self.Base.classes.generic_instance)
            .filter(
                self.Base.classes.generic_instance.category == "actor",
                self.Base.classes.generic_instance.type == "user",
                self.Base.classes.generic_instance.subtype == "bloom-user",
                self.Base.classes.generic_instance.version == "1.0",
                self.Base.classes.generic_instance.is_deleted == False,  # noqa: E712
            )
            .all()
        )

        actor = None
        for candidate in actors:
            _, props = self._instance_properties(candidate)
            candidate_sub = str(props.get("cognito_sub") or "").strip()
            candidate_email = str(props.get("email") or "").strip().lower()
            if resolved_user_id and candidate_sub == resolved_user_id:
                actor = candidate
                break
            if resolved_email and candidate_email == resolved_email.lower():
                actor = candidate
                break

        if actor is None:
            template = self._ensure_user_actor_template()
            actor_name = resolved_email or resolved_user_id or "Bloom User"
            actor = self.create_instance(
                template.euid,
                json_addl_overrides={
                    "properties": {
                        "name": actor_name,
                        "email": resolved_email or "",
                        "cognito_sub": resolved_user_id or "",
                        "comments": "",
                    }
                },
            )
            return actor

        payload, props = self._instance_properties(actor)
        changed = False
        if resolved_email and props.get("email") != resolved_email:
            props["email"] = resolved_email
            changed = True
        if resolved_user_id and props.get("cognito_sub") != resolved_user_id:
            props["cognito_sub"] = resolved_user_id
            changed = True
        if resolved_email and actor.name != resolved_email:
            actor.name = resolved_email
            changed = True
        if changed:
            payload["properties"] = props
            actor.json_addl = payload
            flag_modified(actor, "json_addl")
            self.session.commit()

        return actor

    def track_user_interaction(
        self,
        target_euid,
        *,
        relationship_type="user_interacted",
        action_ds=None,
        user_id=None,
        email=None,
    ):
        """Create lineage from a user actor node to a target object."""
        try:
            target_euid = str(target_euid or "").strip()
            if not target_euid:
                return None

            action_user_id = None
            action_email = None
            if isinstance(action_ds, dict):
                action_user_id, action_email = self._extract_actor_from_curr_user(
                    action_ds.get("curr_user")
                )
                if not action_user_id:
                    action_user_id = (
                        str(action_ds.get("curr_user_id") or "").strip() or None
                    )

            actor = self._upsert_user_actor(
                user_id=user_id or action_user_id or self._actor_user_id,
                email=email or action_email or self._actor_email,
            )
            if actor is None or actor.euid == target_euid:
                return None

            target_obj = self.get_by_euid(target_euid)
            if not target_obj or getattr(target_obj, "is_deleted", False):
                return None

            self.create_generic_instance_lineage_by_euids(
                actor.euid,
                target_euid,
                relationship_type=str(relationship_type or "user_interacted"),
            )
            self.session.commit()
            return actor.euid
        except Exception as exc:
            self.logger.warning(
                "Failed to track user interaction (%s -> %s): %s",
                relationship_type,
                target_euid,
                exc,
            )
            try:
                self.session.rollback()
            except Exception:
                pass
            return None

    def _fetch_fedex_tracking_ops_meta(self, tracking_number):
        """Fetch FedEx metadata across carrier-tracking API versions."""
        if not self.track_fedex:
            return []

        if hasattr(self.track_fedex, "track_ops_meta"):
            data = self.track_fedex.track_ops_meta(tracking_number)
        elif hasattr(self.track_fedex, "get_fedex_ops_meta_ds"):
            data = self.track_fedex.get_fedex_ops_meta_ds(tracking_number)
        elif hasattr(self.track_fedex, "track"):
            data = self.track_fedex.track(tracking_number)
        else:
            raise AttributeError(
                "FedEx tracker supports neither 'track_ops_meta', "
                "'get_fedex_ops_meta_ds', nor 'track'"
            )

        if data is None:
            return []
        if isinstance(data, list):
            return data
        if isinstance(data, dict):
            return [data]
        return [dict(data)]


    def _rebuild_printer_json(self, lab="BLOOM"):
        self.zpld.probe_zebra_printers_add_to_printers_json(lab=lab)
        self.zpld.save_printer_json(self.zpld.printers_filename.split("zebra_day")[-1])
        self._config_printers()

    def _config_printers(self):
        if len(self.zpld.printers["labs"].keys()) == 0:
            self.logger.warning(
                "No printers found, attempting to rebuild printer json\n\n"
            )
            self.logger.warning(
                'This may take a few minutes, lab code will be set to "BLOOM" ... please sit tight...\n\n'
            )
            self._rebuild_printer_json()

        self.printer_labs = self.zpld.printers["labs"].keys()
        self.selected_lab = sorted(self.printer_labs)[0]
        self.site_printers = self.zpld.printers["labs"][self.selected_lab].keys()
        _zpl_label_styles = []
        for zpl_f in os.listdir(
            os.path.dirname(self.zpld.printers_filename) + "/label_styles/"
        ):
            if zpl_f.endswith(".zpl"):
                _zpl_label_styles.append(zpl_f.removesuffix(".zpl"))
        self.zpl_label_styles = sorted(_zpl_label_styles)
        self.selected_label_style = "tube_2inX1in"

    def set_printers_lab(self, lab):
        self.selected_lab = lab

    def get_lab_printers(self, lab):
        self.selected_lab = lab
        try:
            self.site_printers = self.zpld.printers["labs"][self.selected_lab].keys()
        except Exception as e:
            self.logger.error(f"Error getting printers for lab {lab}")
            self.logger.error(e)
            self.logger.error(
                "\n\n\nAttempting to rebuild printer json !!! THIS WILL TAKE TIME !!!\n\n\n"
            )
            self._rebuild_printer_json()

    def print_label(
        self,
        lab=None,
        printer_name=None,
        label_zpl_style="tube_2inX1in",
        euid="",
        alt_a="",
        alt_b="",
        alt_c="",
        alt_d="",
        alt_e="",
        alt_f="",
        print_n=1,
    ):

        bc = self.zpld.print_zpl(
            lab=lab,
            printer_name=printer_name,
            uid_barcode=euid,
            alt_a=alt_a,
            alt_b=alt_b,
            alt_c=alt_c,
            alt_d=alt_d,
            alt_e=alt_e,
            alt_f=alt_f,
            label_zpl_style=label_zpl_style,
            client_ip="pkg",
            print_n=print_n,
        )

    # For use by the cytoscape UI in order to determine if the dag needs regenerating
    def get_most_recent_schema_audit_log_entry(self):
        return (
            self.session.query(self.Base.classes.audit_log)
            .order_by(desc(self.Base.classes.audit_log.changed_at))
            .first()
        )

    # centralizing creation more cleanly.
    def create_instance(self, template_euid, json_addl_overrides={}):
        """Given an EUID for an object template, instantiate an instance from the template.
            No child objects defined by the tmplate will be generated.

            json_addl_overrides is a dict of key value pairs that will be merged into the json_addl of the template, with new keys created and existing keys over written.
        Args:
            template_euid (_type_): _description_
        """

        self.logger.debug(f"Creating instance from template EUID {template_euid}")

        template = self.get_by_euid(template_euid)

        if not template:
            self.logger.debug(f"No template found with euid: " + template_euid)
            return

        is_singleton = (
            False if template.json_addl.get("singleton", "0") in [0, "0"] else True
        )

        # TapDB ships a built-in system user actor template that must be singleton in practice.
        # The underlying DB enforces a uniqueness constraint on `json_addl.login_identifier`
        # for these instances; attempting to create multiple will raise an IntegrityError.
        is_system_user_actor = (
            template.type == "actor"
            and template.subtype == "system_user"
            and str(template.version or "") == "1.0"
        )

        if is_singleton or is_system_user_actor:
            existing = (
                self.session.query(self.Base.classes.generic_instance)
                .filter(
                    self.Base.classes.generic_instance.template_uid == template.uid,
                    self.Base.classes.generic_instance.is_deleted == False,  # noqa: E712
                )
                .first()
            )
            if existing is not None:
                return existing

        cname = template.polymorphic_discriminator.replace("_template", "_instance")
        parent_instance = getattr(self.Base.classes, f"{cname}")(
            name=template.name,
            type=template.type,
            subtype=template.subtype,
            version=template.version,
            json_addl=template.json_addl,
            template_uid=template.uid,
            bstatus=template.bstatus,
            category=template.category,
            is_singleton=is_singleton,
            polymorphic_discriminator=template.polymorphic_discriminator.replace(
                "_template", "_instance"
            ),
        )

        # Normalize TapDB system-user actor instances to satisfy DB uniqueness constraint.
        if is_system_user_actor:
            # Promote properties.login_identifier to top-level json_addl.login_identifier.
            try:
                payload = parent_instance.json_addl or {}
                props = payload.get("properties") or {}
                login_identifier = str(payload.get("login_identifier") or "").strip()
                if not login_identifier:
                    login_identifier = str(props.get("login_identifier") or "").strip()
                if not login_identifier:
                    login_identifier = "system"
                payload["login_identifier"] = login_identifier
                parent_instance.json_addl = payload
            except Exception:
                # Best-effort: if anything goes wrong, let the DB enforce constraints.
                pass

        # Lots of fun stuff happening when instantiating action_imports!
        ai = (
            self._create_action_ds(parent_instance.json_addl["action_imports"])
            if "action_imports" in parent_instance.json_addl
            else {}
        )
        try:
            json_addl_overrides["action_groups"] = ai
            _update_recursive(parent_instance.json_addl, json_addl_overrides)
            self.session.add(parent_instance)
            ##self.session.flush()
            self.session.commit()
        except Exception as e:
            self.logger.error(f"Error creating instance from template {template_euid}")
            self.logger.error(e)
            self.session.rollback()
            raise Exception(
                f"Error creating instance from template {template_euid} ... {e} .. Likely Singleton Violation"
            )

        return parent_instance
    # fix naming, instance_type==table_name_prefix
    def create_instances(self, template_euid):
        """
        IMPORTANTLY: this method creates the requested object from the template, and also will recurse one level to create any children objects defined by the template.
        You get back an array with the first element being the parent the second an array of children.

        Create instances from exsiting templates in the *_template table that the euid is a member of.
        The class subclassing is an experiment, see the docs (hopefully) for more, in short:
            TABLE_template entries hold the TABLE_instance definitions (and to a large extend, the TABLE_instance_lineage as well)
            TABLE_template is seeded initially from json files in
            bloom_lims/config/{container,content,workflow,workflow_step,equipment,data,test_requisition,...}/*.json
            These are only used to seed the templates... the idea is to allow users to add subtypes as they see fit. (TBD)

            ~ TABLE_template/{type}/{subtype}/{version} is the pattern used to for the template table name
        This is a recursive function that will only create one level of children instances.

        Args:
            template_euid (_type_): a template euid of the pattern [A-Z][A-Z]T[0-9]+ , which is a nicety and nothing at all is ever inferred from the prefix.
            For more on enterprise uuids see: my rants, and (need to find stripe primary ref: https://clerk.com/blog/generating-sortable-stripe-like-ids-with-segment-ksuids)

        Returns:
            [[],[]]: arr[0][:] are parents (presently, there is only ever 1 parent), arr[1][:] are children, if any.
        """

        self.logger.debug(f"Creating instances from template EUID {template_euid}")
        template = self.get_by_euid(
            template_euid
        )  # needed to get this for the child recrods if any
        parent_instance = self.create_instance(template_euid)
        ret_objs = [[], []]
        ret_objs[0].append(parent_instance)

        # template_json_addl = template.json_addl
        if "instantiation_layouts" in template.json_addl:
            ret_objs = self._process_instantiation_layouts(
                template.json_addl["instantiation_layouts"],
                parent_instance,
                ret_objs,
            )
        ##self.session.flush()
        self.session.commit()

        return ret_objs

    # I am of two minds re: if actions should be full objects, or pseudo-objects as they are now...
    def _create_action_ds(self, action_imports):
        ret_ds = {}

        if not isinstance(action_imports, dict):
            return ret_ds

        normalized_groups = action_imports

        # TapDB action imports can be a flat map: {name: "category/type/subtype/version"}.
        # Normalize to Bloom's grouped format for downstream processing.
        if all(isinstance(v, str) for v in action_imports.values()):
            flat_actions = {}
            for template_code in action_imports.values():
                code = template_code if str(template_code).endswith("/") else f"{template_code}/"
                flat_actions[code] = {}
            normalized_groups = {
                "core": {
                    "group_order": "1",
                    "group_name": "Core Actions",
                    "actions": flat_actions,
                }
            }

        for group in normalized_groups:
            group_cfg = normalized_groups.get(group, {})
            if not isinstance(group_cfg, dict):
                continue

            actions_cfg = group_cfg.get("actions", {})
            if not isinstance(actions_cfg, dict):
                continue

            ret_ds[group] = {}
            ret_ds[group]["actions"] = {}
            ret_ds[group]["group_order"] = str(group_cfg.get("group_order", "1"))
            ret_ds[group]["group_name"] = str(group_cfg.get("group_name", group))
            for ai in actions_cfg:
                sl = ai.lstrip("/").split("/")
                category = None if sl[0] == "*" else sl[0]
                type_name = None if sl[1] == "*" else sl[1]
                subtype = None if sl[2] == "*" else sl[2]
                version = None if sl[3] == "*" else sl[3]

                res = self.query_template_by_component_v2(
                    category, type_name, subtype, version
                )
                if len(res) == 0:
                    raise Exception(f"Action import {ai} not found in database")

                for r in res:
                    action_key = f"{r.category}/{r.type}/{r.subtype}/{r.version}"
                    action_payload = None
                    if isinstance(r.json_addl, dict):
                        action_payload = r.json_addl.get("action_definition")
                    if action_payload is None:
                        self.logger.debug(
                            "Ignoring retired action import without action_definition: %s",
                            action_key,
                        )
                        continue

                    ret_ds[group]["actions"][action_key] = {
                        **action_payload,
                        "action_template_uid": str(r.uid),
                        "action_template_euid": r.euid,
                        "action_template_code": action_key,
                    }

                    #  I'm allowing overrides to the action properties FROM
                    # The non-action object action definition.  Its mostly shaky b/c the overrides are applied to all actions
                    # in the matched group... so, when all core are imported for example, an override will match all
                    # I think...  for singleton imports should be ok.
                    # this is to be used mostly for the assay links for test requisitions
                    _update_recursive(
                        ret_ds[group]["actions"][action_key],
                        actions_cfg.get(ai, {}),
                    )

        return ret_ds

    def _process_instantiation_layouts(
        self,
        instantiation_layouts,
        parent_instance,
        ret_objs,
    ):
        # Revisit the lineage set creation, this will not behave as expected if the json templates define more than 1 level deep children.
        ## or is this desireable, and the referenced children should reference thier children... crazy town begins at this level...
        for row in instantiation_layouts:
            for ds in row:
                for i in ds:
                    layout_str = i
                    layout_ds = ds[i]
                    child_instance = self._create_child_instance(layout_str, layout_ds)
                    lineage_record = self.Base.classes.generic_instance_lineage(
                        parent_instance_uid=parent_instance.uid,
                        child_instance_uid=child_instance.uid,
                        name=f"{parent_instance.name} :: {child_instance.name}",
                        type=parent_instance.type,
                        subtype=parent_instance.subtype,
                        version=parent_instance.version,
                        json_addl=parent_instance.json_addl,
                        bstatus=parent_instance.bstatus,
                        category=parent_instance.category,
                        parent_type=parent_instance.polymorphic_discriminator,
                        child_type=child_instance.polymorphic_discriminator,
                        polymorphic_discriminator=f"{parent_instance.category}_instance_lineage",
                    )
                    self.session.add(lineage_record)
                    ##self.session.flush()
                    ret_objs[1].append(child_instance)

        return ret_objs

    def create_generic_instance_lineage_by_euids(
        self, parent_instance_euid, child_instance_euid, relationship_type="generic"
    ):
        parent_instance = self.get_by_euid(parent_instance_euid)
        child_instance = self.get_by_euid(child_instance_euid)
        lineage_record = self.Base.classes.generic_instance_lineage(
            parent_instance_uid=parent_instance.uid,
            child_instance_uid=child_instance.uid,
            name=f"{parent_instance.name} :: {child_instance.name}",
            type=parent_instance.type,
            subtype=parent_instance.subtype,
            version=parent_instance.version,
            json_addl=parent_instance.json_addl,
            bstatus=parent_instance.bstatus,
            category="generic",
            parent_type=f"{parent_instance.category}:{parent_instance.type}:{parent_instance.subtype}:{parent_instance.version}",
            child_type=f"{child_instance.category}:{child_instance.type}:{child_instance.subtype}:{child_instance.version}",
            polymorphic_discriminator=f"generic_instance_lineage",
            relationship_type=relationship_type,
        )
        self.session.add(lineage_record)
        self.session.flush()
        # self.session.commit()

        return lineage_record

    def create_lineage(self, parent_instance, child_instance, relationship_type: str = "generic"):
        """
        Backwards-compatible lineage creator.

        Many API routes and legacy code paths refer to `create_lineage(parent, child)`.
        TapDB/Bloom core implementation is `create_generic_instance_lineage_by_euids`.
        """
        parent_euid = getattr(parent_instance, "euid", parent_instance)
        child_euid = getattr(child_instance, "euid", child_instance)
        if not parent_euid or not child_euid:
            raise Exception("parent_instance and child_instance must be provided")
        return self.create_generic_instance_lineage_by_euids(
            parent_instance_euid=parent_euid,
            child_instance_euid=child_euid,
            relationship_type=relationship_type,
        )

    def create_instance_by_code(self, layout_str, layout_ds):
        ret_obj = self._create_child_instance(layout_str, layout_ds)

        return ret_obj

    def _create_child_instance(self, layout_str, layout_ds):
        (
            category,
            type_name,
            subtype,
            version,
            defaults,
        ) = self._parse_layout_string(layout_str)

        defaults_ds = {}
        ## is this supposed to be coming from the defaults arg above??? hmmm

        if "json_addl" in layout_ds:
            defaults_ds = layout_ds["json_addl"]

        # Not implementing now, assuming all are 1.0
        ## !!! I THINK * IS A POOR IDEA NOW.... considering * to == 1.0 now
        if version == "*":
            version = "1.0"

        templates = self.query_template_by_component_v2(
            category, type_name, subtype, version
        )
        if not templates:
            raise Exception(
                f"Template not found: {category}/{type_name}/{subtype}/{version}. "
                "Please ensure the database is seeded with templates."
            )
        template = templates[0]

        new_instance = self.create_instance(template.euid)
        _update_recursive(new_instance.json_addl, defaults_ds)
        flag_modified(new_instance, "json_addl")
        ##self.session.flush()
        self.session.commit()

        return new_instance

    def _parse_layout_string(self, layout_str):
        parts = layout_str.split("/")
        category = parts[0]  # table name now called 'category'
        type_name = parts[1] if len(parts) > 1 else "*"
        subtype = parts[2] if len(parts) > 2 else "*"
        version = (
            parts[3] if len(parts) > 3 else "*"
        )  # Assuming the version is always the third part
        defaults = (
            parts[4] if len(parts) > 4 else ""
        )  # Assuming the defaults is always the fourth part
        return category, type_name, subtype, version, defaults

    # json additional information validators
    def validate_object_vs_pattern(self, obj, pattern):
        """
        Validates if a given object matches the given pattern

        Args:
            obj _type_: _description_
            pattern _type_: _description_

        Returns:
            _type_: bool()
        """

        # Parse the JSON additional information of the object
        classn = (
            str(obj.__class__)
            .split(".")[-1]
            .replace(">", "")
            .replace("_instance", "")
            .replace("'", "")
        )

        obj_type_info = f"{classn}/{obj.type}/{obj.subtype}/{obj.version}"

        # Check if the object matches the pattern
        compiled_pattern = re.compile(pattern)
        match = compiled_pattern.search(obj_type_info)

        if match:
            return True

        return False

    """Lookup methods. EUID is the canonical Bloom object identifier."""

    def get_by_euid(self, euid):
        """Global query for euid across all tables in schema with 'euid' field
           note: does not handle is_deleted!
        Args:
            euid str(): euid string

        Returns:
            [] : Array of rows
        """
        if euid is None:
            raise Exception("euid cannot be None")
        res = (
            self.session.query(self.Base.classes.generic_instance)
            .filter(
                self.Base.classes.generic_instance.euid == euid,
                self.Base.classes.generic_instance.is_deleted == self.is_deleted,
            )
            .all()
        )
        res2 = (
            self.session.query(self.Base.classes.generic_template)
            .filter(
                self.Base.classes.generic_template.euid == euid,
                self.Base.classes.generic_template.is_deleted == self.is_deleted,
            )
            .all()
        )
        res3 = (
            self.session.query(self.Base.classes.generic_instance_lineage)
            .filter(
                self.Base.classes.generic_instance_lineage.euid == euid,
                self.Base.classes.generic_instance_lineage.is_deleted
                == self.is_deleted,
            )
            .all()
        )

        combined_result = res + res2 + res3

        if len(combined_result) > 1:
            raise Exception(
                f"Multiple {len(combined_result)} templates found for {euid}"
            )
        elif len(combined_result) == 0:
            self.logger.debug(f"No template found with euid: " + euid)
            raise Exception(f"No template found with euid: " + euid)
        else:
            return combined_result[0]

    # This is the mechanism for finding the database object(s) which match the template reference pattern
    # V2... why?
    def query_instance_by_component_v2(
        self, category=None, type=None, subtype=None, version=None
    ):
        query = self.session.query(self.Base.classes.generic_instance)

        # Apply filters conditionally
        if category is not None:
            query = query.filter(
                self.Base.classes.generic_instance.category == category
            )
        if type is not None:
            query = query.filter(self.Base.classes.generic_instance.type == type)
        if subtype is not None:
            query = query.filter(
                self.Base.classes.generic_instance.subtype == subtype
            )
        if version is not None:
            query = query.filter(self.Base.classes.generic_instance.version == version)

        query = query.filter(
            self.Base.classes.generic_instance.is_deleted == self.is_deleted
        )

        # Execute the query
        return query.all()


    # should abstract to not assume properties key
    def get_unique_property_values(self, property_key, category=None, type=None, subtype=None, version=None):

        json_path = property_key.split("->")

        # Start building the query with the base table
        query = self.session.query(
            self.Base.classes.generic_instance
        )

        # Add filters based on the provided arguments
        if category:
            query = query.filter(self.Base.classes.generic_instance.category == category)
        if type:
            query = query.filter(self.Base.classes.generic_instance.type == type)
        if subtype:
            query = query.filter(self.Base.classes.generic_instance.subtype == subtype)
        if version:
            query = query.filter(self.Base.classes.generic_instance.version == version)

        # Add the JSON path extraction and distinct filtering after the base filters
        query = query.with_entities(
            func.distinct(
                func.jsonb_extract_path_text(
                    self.Base.classes.generic_instance.json_addl['properties'], *json_path
                )
            )
        ).filter(
            func.jsonb_extract_path_text(
                self.Base.classes.generic_instance.json_addl['properties'], *json_path
            ).isnot(None)
        )

        # Execute the query and get the results
        results = query.all()

        # Extract unique values from the query results
        unique_values = [value[0] for value in results if value[0] is not None]
        
        return unique_values


    def query_template_by_component_v2(
        self, category=None, type=None, subtype=None, version=None
    ):
        query = self.session.query(self.Base.classes.generic_template)

        # Apply filters conditionally
        if category is not None:
            query = query.filter(
                self.Base.classes.generic_template.category == category
            )
        if type is not None:
            query = query.filter(self.Base.classes.generic_template.type == type)

        if subtype is not None:
            query = query.filter(
                self.Base.classes.generic_template.subtype == subtype
            )
        if version is not None:
            query = query.filter(self.Base.classes.generic_template.version == version)

        query = query.filter(
            self.Base.classes.generic_template.is_deleted == self.is_deleted
        )
        # Execute the query
        return query.all()

   
    def query_user_audit_logs(self, username):
        logging.debug(f"Querying audit log for user: {username}")

        q = text(
            """
            SELECT
                al.rel_table_euid_fk AS euid,
                al.changed_by,
                al.operation_type,
                al.changed_at,
                COALESCE(gt.name, gi.name, gil.name) AS name,
                COALESCE(gt.polymorphic_discriminator, gi.polymorphic_discriminator, gil.polymorphic_discriminator) AS polymorphic_discriminator,
                COALESCE(gt.category, gi.category, gil.category) AS category,
                COALESCE(gt.type, gi.type, gil.type) AS type,
                COALESCE(gt.subtype, gi.subtype, gil.subtype) AS subtype,
                COALESCE(gt.bstatus, gi.bstatus, gil.bstatus) AS status,
                al.old_value,
                al.new_value
            FROM
                audit_log al
                LEFT JOIN generic_template gt ON al.rel_table_uid_fk = gt.uid
                LEFT JOIN generic_instance gi ON al.rel_table_uid_fk = gi.uid
                LEFT JOIN generic_instance_lineage gil ON al.rel_table_uid_fk = gil.uid
            WHERE
                al.changed_by = :username
            ORDER BY
                al.changed_at DESC;
            """
        )

        logging.debug(f"Executing query: {q}")

        result = self.session.execute(q, {'username': username})
        rows = result.fetchall()

        logging.debug(f"Query returned {len(rows)} rows")

        return rows
    # Aggregate Report SQL
    def query_generic_template_stats(self):
        q = text(
            """
            SELECT
                'Generic Template Summary' as Report,
                COUNT(*) as Total_Templates,
                COUNT(DISTINCT type) as Distinct_Base_Types,
                COUNT(DISTINCT subtype) as Distinct_Sub_Types,
                COUNT(DISTINCT category) as Distinct_Super_Types,
                MAX(created_dt) as Latest_Creation_Date,
                MIN(created_dt) as Earliest_Creation_Date,
                AVG(AGE(NOW(), created_dt)) as Average_Age,
                COUNT(CASE WHEN is_singleton THEN 1 END) as Singleton_Count
            FROM
                generic_template
            WHERE
                is_deleted = :is_deleted
        """
        )

        result = self.session.execute(q, {"is_deleted": self.is_deleted}).fetchall()

        # Define the column names based on your SELECT statement
        columns = [
            "Report",
            "Total_Templates",
            "Distinct_Base_Types",
            "Distinct_Sub_Types",
            "Distinct_Super_Types",
            "Latest_Creation_Date",
            "Earliest_Creation_Date",
            "Average_Age",
            "Singleton_Count",
        ]

        # Convert each row to a dictionary
        return [dict(zip(columns, row)) for row in result]

    def query_generic_instance_and_lin_stats(self):
        q = text(
            f"""
        SELECT
            -- Summary from generic_instance table
            'Generic Instance Summary' as Report,
            COUNT(*) as Total_Instances,
            COUNT(DISTINCT type) as Distinct_Types,
            COUNT(DISTINCT polymorphic_discriminator) as Distinct_Polymorphic_Discriminators,
            COUNT(DISTINCT category) as Distinct_Super_Types,
            COUNT(DISTINCT subtype) as Distinct_Sub_Types,
            MAX(created_dt) as Latest_Creation_Date,
            MIN(created_dt) as Earliest_Creation_Date,
            AVG(AGE(NOW(), created_dt)) as Average_Age
        FROM
            generic_instance
        WHERE
            is_deleted = {self.is_deleted}

        UNION ALL

        SELECT
            -- Summary from generic_instance_lineage table
            'Generic Instance Lineage Summary',
            COUNT(*) as Total_Lineages,
            COUNT(DISTINCT parent_type) as Distinct_Parent_Types,
            COUNT(DISTINCT child_type) as Distinct_Child_Types,
            COUNT(DISTINCT polymorphic_discriminator) as Distinct_Polymorphic_Discriminators,
            COUNT(DISTINCT category) as Distinct_Super_Types,
            MAX(created_dt) as Latest_Creation_Date,
            MIN(created_dt) as Earliest_Creation_Date,
            AVG(AGE(NOW(), created_dt)) as Average_Age
        FROM
            generic_instance_lineage
        WHERE
            is_deleted = {self.is_deleted};
        """
        )

        result = self.session.execute(q, {"is_deleted": self.is_deleted}).fetchall()

        # Define the column names based on your SELECT statement
        columns = [
            "Report",
            "Total_Instances",
            "Distinct_Types",
            "Distinct_Polymorphic_Discriminators",
            "Distinct_Super_Types",
            "Distinct_Sub_Types",
            "Latest_Creation_Date",
            "Earliest_Creation_Date",
            "Average_Age",
        ]

        # Convert each row to a dictionary
        return [dict(zip(columns, row)) for row in result]

    def query_cost_of_all_children(self, euid):
        # limited to 10,000 children right now...
        query = text(
            """
            WITH RECURSIVE descendants AS (
            -- Initial query to get the root instance
            SELECT gi.uid, gi.euid, gi.json_addl, gi.created_dt
            FROM generic_instance gi
            WHERE gi.euid = :euid

            UNION ALL

            -- Recursive part to get all descendants
            SELECT child_gi.uid, child_gi.euid, child_gi.json_addl, child_gi.created_dt
            FROM generic_instance_lineage gil
            JOIN descendants d ON gil.parent_instance_uid = d.uid
            JOIN generic_instance child_gi ON gil.child_instance_uid = child_gi.uid
            WHERE NOT child_gi.is_deleted -- Assuming you want to exclude deleted instances
        )
        SELECT d.euid, 
            d.json_addl -> 'cogs' ->> 'cost' AS cost
        FROM descendants d
        WHERE d.json_addl ? 'cogs' AND 
            d.json_addl -> 'cogs' ? 'cost' AND 
            d.json_addl -> 'cogs' ->> 'cost' <> ''
        ORDER BY d.created_dt DESC -- Order the final result set        
        """
        )

        # Execute the query
        result = self.session.execute(query, {"euid": str(euid)})

        # Extract euids and transit times from the result
        euid_cost_tuples = [(row[0], row[1]) for row in result]

        return euid_cost_tuples

    def query_all_fedex_transit_times_by_ay_euid(self, qx_euid):

        query = text(
            """SELECT gi.euid,
        gi.json_addl -> 'properties' -> 'fedex_tracking_data' -> 0 ->> 'Transit_Time_sec' AS transit_time
        FROM generic_instance AS gi
        JOIN generic_instance_lineage AS gil1 ON gi.uid = gil1.child_instance_uid
        JOIN generic_instance AS gi_parent1 ON gil1.parent_instance_uid = gi_parent1.uid
        JOIN generic_instance_lineage AS gil2 ON gi_parent1.uid = gil2.child_instance_uid
        JOIN generic_instance AS gi_parent2 ON gil2.parent_instance_uid = gi_parent2.uid
        WHERE
        gi_parent2.euid = :qx_euid AND
        gi.type = 'package' AND
        jsonb_typeof(gi.json_addl -> 'properties') = 'object' AND
        jsonb_typeof(gi.json_addl -> 'properties' -> 'fedex_tracking_data') = 'array' AND
        jsonb_typeof((gi.json_addl -> 'properties' -> 'fedex_tracking_data' -> 0)) = 'object' AND
        COALESCE(NULLIF(gi.json_addl -> 'properties' -> 'fedex_tracking_data' -> 0 ->> 'Transit_Time_sec', ''), '0') >= '0';
        """
        )

        # Execute the query
        result = self.session.execute(query, {"qx_euid": qx_euid})

        # Extract euids and transit times from the result
        euid_transit_time_tuples = [(row[0], row[1]) for row in result]

        return euid_transit_time_tuples

    def fetch_graph_data_by_node_depth(self, start_euid, depth):
        """
        Fetch graph data for a node and its neighbors up to a specified depth.

        Uses parameterized queries to prevent SQL injection attacks.

        Args:
            start_euid: The EUID of the starting node
            depth: Maximum depth to traverse from the starting node

        Returns:
            Query result containing node and edge data
        """
        # SQL query with parameterized placeholders for security
        # Note: Uses TapDB column names (type, category, subtype) directly
        query = text(
            """WITH RECURSIVE graph_data AS (
                SELECT
                    gi.euid,
                    gi.uid,
                    gi.name,
                    gi.type,
                    gi.category,
                    gi.subtype,
                    gi.version,
                    0 AS depth,
                    NULL::text AS lineage_euid,
                    NULL::text AS lineage_parent_euid,
                    NULL::text AS lineage_child_euid,
                    NULL::text AS relationship_type
                FROM
                    generic_instance gi
                WHERE
                    gi.euid = :start_euid AND gi.is_deleted = FALSE

                UNION

                SELECT
                    gi.euid,
                    gi.uid,
                    gi.name,
                    gi.type,
                    gi.category,
                    gi.subtype,
                    gi.version,
                    gd.depth + 1,
                    gil.euid AS lineage_euid,
                    parent_instance.euid AS lineage_parent_euid,
                    child_instance.euid as lineage_child_euid,
                    gil.relationship_type
                FROM
                    generic_instance_lineage gil
                JOIN
                    generic_instance gi ON gi.uid = gil.child_instance_uid OR gi.uid = gil.parent_instance_uid
                JOIN
                    generic_instance parent_instance ON gil.parent_instance_uid = parent_instance.uid
                JOIN
                    generic_instance child_instance ON gil.child_instance_uid = child_instance.uid
                JOIN
                    graph_data gd ON (gil.parent_instance_uid = gd.uid AND gi.uid = gil.child_instance_uid) OR
                                    (gil.child_instance_uid = gd.uid AND gi.uid = gil.parent_instance_uid)
                WHERE
                    gi.is_deleted = FALSE AND gil.is_deleted = FALSE AND gd.depth < :depth
            )
            SELECT DISTINCT * FROM graph_data;
        """
        )

        # Execute the query with bound parameters (prevents SQL injection)
        result = self.session.execute(
            query,
            {"start_euid": str(start_euid), "depth": int(depth)}
        )
        return result

    def create_instance_by_template_components(
        self, category, type, subtype, version
    ):
        templates = self.query_template_by_component_v2(category, type, subtype, version)
        if not templates:
            raise Exception(
                f"Template not found: {category}/{type}/{subtype}/{version}. "
                "Please ensure the database is seeded with templates (run: bloom init)."
            )
        return self.create_instances(templates[0].euid)

    # Is this too special casey? Belong lower?
    def create_container_with_content(self, cx_quad_tup, mx_quad_tup):
        """ie CX=container, MX=content (material)
        ("content", "control", "giab-HG002", "1.0"),
        ("container", "tube", "tube-generic-10ml", "1.0")
        """
        cx_templates = self.query_template_by_component_v2(
            cx_quad_tup[0], cx_quad_tup[1], cx_quad_tup[2], cx_quad_tup[3]
        )
        if not cx_templates:
            raise Exception(
                f"Container template not found: {'/'.join(cx_quad_tup)}. "
                "Please ensure the database is seeded with templates."
            )
        container = self.create_instance(cx_templates[0].euid)

        mx_templates = self.query_template_by_component_v2(
            mx_quad_tup[0], mx_quad_tup[1], mx_quad_tup[2], mx_quad_tup[3]
        )
        if not mx_templates:
            raise Exception(
                f"Content template not found: {'/'.join(mx_quad_tup)}. "
                "Please ensure the database is seeded with templates."
            )
        content = self.create_instance(mx_templates[0].euid)

        container.json_addl["properties"]["name"] = content.json_addl["properties"][
            "name"
        ]
        flag_modified(container, "json_addl")
        ##self.session.flush()
        self.create_generic_instance_lineage_by_euids(container.euid, content.euid)
        self.session.commit()

        return container, content

    # Delete Methods
    # Do not cascade delete!

    def delete(self, euid):
        if euid is None:
            raise Exception("euid cannot be None")
        obj = self.get_by_euid(euid)
        obj.is_deleted = True
        ##self.session.flush()
        self.session.commit()

    def delete_by_euid(self, euid):
        return self.delete(euid)

    def delete_obj(self, obj):
        if obj is None:
            raise Exception("obj cannot be None")
        obj.is_deleted = True
        ##self.session.flush()
        self.session.commit()

    def do_action_add_file_to_file_set(self, file_set_euid, action_ds):
        bfs = BloomFileSet(BLOOMdb3())
        bfs.add_files_to_file_set(
            euid=file_set_euid, file_euid=[action_ds["captured_data"]["file_euid"]]
        )

    def do_action_remove_file_from_file_set(self, file_set_euid, action_ds):
        bfs = BloomFileSet(BLOOMdb3())
        bfs.remove_files_from_file_set(
            euid=file_set_euid, file_euid=[action_ds["captured_data"]["file_euid"]]
        )

    def do_action_add_relationships(self, euid, action_ds):

        euid_obj = self.get_by_euid(euid)
        lineage_to_create = action_ds["captured_data"]["lineage_type_to_create"]
        relationship_type = action_ds["captured_data"]["relationship_type"]
        euids = action_ds["captured_data"]["euids"]

        # euids is the text from a textareas, process each and assign lineage
        for a_euid in euids.split("\n"):
            if a_euid != "":
                if lineage_to_create == "parent":
                    self.create_generic_instance_lineage_by_euids(
                        a_euid, euid, relationship_type
                    )
                elif lineage_to_create == "child":
                    self.create_generic_instance_lineage_by_euids(
                        euid, a_euid, relationship_type
                    )
                else:
                    self.logger.exception(
                        f"Unknown lineage type {lineage_to_create}, requires 'parent' or 'child'"
                    )
                    raise Exception(
                        f"Unknown lineage type {lineage_to_create}, requires 'parent' or 'child'"
                    )

        return euid_obj

    def do_action_create_subject_and_anchor(self, euid, action_ds):
        """
        Create a Subject (decision scope) with the given object as the anchor.

        Args:
            euid: EUID of the object to use as anchor
            action_ds: Action data structure with captured_data containing:
                - subject_kind: Type of subject (accession, analysis_bundle, report, generic)
                - subject_name: Optional name for the subject
                - comments: Optional comments

        Returns:
            The created subject EUID, or None if creation failed
        """
        from bloom_lims.subjecting import create_subject

        captured = action_ds.get("captured_data", {})
        subject_kind = captured.get("subject_kind", "generic")
        subject_name = captured.get("subject_name", "")
        comments = captured.get("comments", "")

        extra_props = {}
        if subject_name:
            extra_props["name"] = subject_name
        if comments:
            extra_props["comments"] = comments

        subject_euid = create_subject(
            bob=self,
            anchor_euid=euid,
            subject_kind=subject_kind,
            extra_props=extra_props,
        )

        if subject_euid:
            self.logger.info(f"Created subject {subject_euid} for anchor {euid}")
        else:
            self.logger.error(f"Failed to create subject for anchor {euid}")

        return subject_euid

    def ret_plate_wells_dict(self, plate):
        plate_wells = {}
        for lin in plate.parent_of_lineages:
            if lin.child_instance.type == "well":

                well = lin.child_instance
                content_arr = []
                for c in well.parent_of_lineages:
                    if c.child_instance.category == "content":
                        content_arr.append(c.child_instance)
                content = None
                if len(content_arr) == 0:
                    pass
                elif len(content_arr) == 1:
                    content = content_arr[0]
                else:
                    self.logger.exception(
                        f"More than one content found for well {well.euid}"
                    )
                    raise Exception(f"More than one content found for well {well.euid}")

                plate_wells[lin.child_instance.json_addl["cont_address"]["name"]] = (
                    lin.child_instance,
                    content,
                )

        return plate_wells

    def do_action_download_file(self, euid, action_ds):

        bf = BloomFile(BLOOMdb3())
        dl_file = bf.download_file(
            euid=euid,
            include_metadata=(
                True
                if action_ds["captured_data"]["create_metadata_file"] in ["yes"]
                else False
            ),
            save_path="./tmp/",
            save_pattern=action_ds["captured_data"]["download_type"],
        )
        return dl_file

    def do_stamp_plates_into_plate(self, euid, action_ds):
        # Taking a stab at moving to a non obsessive commit world

        euid_obj = self.get_by_euid(euid)

        dest_plate = self.get_by_euid(
            action_ds["captured_data"]["Destination Plate EUID"]
        )
        source_plates = []
        source_plates_well_digested = []
        for source_plt_euid in action_ds["captured_data"]["source_barcodes"].split(
            "\n"
        ):
            spo = self.get_by_euid(source_plt_euid)
            source_plates.append(spo)
            source_plates_well_digested.append(self.ret_plate_wells_dict(spo))

        wfs = ""
        for layout_str in action_ds["child_workflow_step_obj"]:
            wfs = self.create_instance_by_code(
                layout_str, action_ds["child_workflow_step_obj"][layout_str]
            )
            self.create_generic_instance_lineage_by_euids(euid_obj.euid, wfs.euid)

        self.create_generic_instance_lineage_by_euids(wfs.euid, dest_plate.euid)

        for spo in source_plates:
            self.create_generic_instance_lineage_by_euids(wfs.euid, spo.euid)

        # For all plates being stamped into the destination, link all source plate wells to the destination plate wells, and the contensts of source wells to destination wells.
        # Further, if a dest well is empty, create a new content instance for it and link appropriately.
        for dest_well in dest_plate.parent_of_lineages:
            if dest_well.child_instance.type == "well":
                well_name = dest_well.child_instance.json_addl["cont_address"]["name"]
                for spod in source_plates_well_digested:
                    if well_name in spod:
                        self.create_generic_instance_lineage_by_euids(
                            spod[well_name][0].euid, dest_well.child_instance.euid
                        )
                        if spod[well_name][1] != None:
                            for dwc in dest_well.child_instance.parent_of_lineages:
                                if dwc.child_instance.category == "content":
                                    self.create_generic_instance_lineage_by_euids(
                                        spod[well_name][1].euid, dwc.child_instance.euid
                                    )
                        del spod[well_name]
        ## TODO
        ### IF there are any source wells left, create new content instances for them and link to the dest wells
        remaining_wells = 0
        for i in source_plates_well_digested:
            for j in i:
                remaining_wells += 1
        if remaining_wells > 0:
            self.logger.exception(
                f"ERROR: {remaining_wells} wells left over after stamping"
            )
            self.session.rollback()
            raise Exception(f"ERROR: {remaining_wells} wells left over after stamping")

        self.session.commit()

        return wfs

    def do_action_move_workset_to_another_queue(self, euid, action_ds):

        wfset = self.get_by_euid(euid)
        action_ds["captured_data"]["q_selection"]

        # Filter to only get active (non-deleted) lineages
        active_child_of_lineages = [lin for lin in wfset.child_of_lineages if not lin.is_deleted]

        # EXTRAORDINARILY SLOPPY.  I AM IN A REAL RUSH FOR FEATURES THO :-/
        destination_q = ""
        (category, type_name, subtype, version) = (
            action_ds["captured_data"]["q_selection"].lstrip("/").rstrip("/").split("/")
        )

        if len(active_child_of_lineages) == 0:
            self.logger.exception(f"ERROR: No active child_of_lineages for {euid}")
            raise Exception(f"ERROR: No active child_of_lineages for workset {euid}")

        # Get active lineages for traversal (filter out deleted ones)
        current_lineage = active_child_of_lineages[0]
        parent_active_lineages = [lin for lin in current_lineage.parent_instance.child_of_lineages if not lin.is_deleted]

        if len(parent_active_lineages) == 0:
            self.logger.exception(f"ERROR: No active parent lineages found")
            raise Exception(f"ERROR: No active parent lineages found for queue traversal")

        parent_of_parent = parent_active_lineages[0].parent_instance
        for q in parent_of_parent.parent_of_lineages:
            if q.is_deleted:
                continue
            if (
                q.child_instance.type == type_name
                and q.child_instance.subtype == subtype
            ):
                destination_q = q.child_instance
                break

        if len(active_child_of_lineages) != 1 or destination_q == "":
            self.logger.exception(f"ERROR: {action_ds['captured_data']['q_selection']} - active lineages: {len(active_child_of_lineages)}, destination_q found: {destination_q != ''}")
            raise Exception(f"ERROR: {action_ds['captured_data']['q_selection']}")

        lineage_link = active_child_of_lineages[0]
        self.create_generic_instance_lineage_by_euids(destination_q.euid, wfset.euid)
        self.delete_obj(lineage_link)
        ##self.session.flush()
        self.session.commit()

    # Doing this globally for now
    def do_action_create_package_and_first_workflow_step_assay(
        self, euid, action_ds={}
    ):
        raise RuntimeError(
            "Package/kit accessioning actions have been retired from Bloom. "
            "Accessioning ownership is Atlas-only."
        )

    def do_action_print_barcode_label(self, euid, action_ds={}):
        """_summary_

        Args:
            euid (str()): bloom obj EUID
            action (str()): action name from object json_addl['actions']
            action_ds (dict): the dictionary keyed by the object json_addl['action'][action]
        """
        bobj = self.get_by_euid(euid)

        lab = action_ds.get("lab", "")
        printer_name = action_ds.get("printer_name", "")
        label_zpl_style = action_ds.get("label_style", "")
        alt_a = (
            action_ds.get("alt_a", "")
            if not PGLOBAL
            else f"{bobj.subtype}-{bobj.version}"
        )
        alt_b = (
            action_ds.get("alt_b", "")
            if not PGLOBAL
            else bobj.json_addl.get("properties", {}).get("name", "__namehere__")
        )
        alt_c = (
            action_ds.get("alt_c", "")
            if not PGLOBAL
            else bobj.json_addl.get("properties", {}).get("lab_code", "N/A")
        )
        alt_d = action_ds.get("alt_d", "")
        alt_e = (
            action_ds.get("alt_e", "")
            if not PGLOBAL
            else str(bobj.created_dt).split(" ")[0]
        )
        alt_f = action_ds.get("alt_f", "")

        self.logger.info(
            f"PRINTING BARCODE LABEL for {euid} at {lab} .. {printer_name} .. {label_zpl_style} \n"
        )

        self.print_label(
            lab=lab,
            printer_name=printer_name,
            label_zpl_style=label_zpl_style,
            euid=euid,
            alt_a=alt_a,
            alt_b=alt_b,
            alt_c=alt_c,
            alt_d=alt_d,
            alt_e=alt_e,
            alt_f=alt_f,
        )

    def do_action_set_object_status(
        self, euid, action_ds={}, action_group=None, action=None
    ):
        bobj = self.get_by_euid(euid)

        now_dt = get_datetime_string()
        un = action_ds.get("curr_user", "bloomdborm")
        status = action_ds["captured_data"]["object_status"]
        try:
            if status == "in_progress":
                if bobj.bstatus in ["complete", "abandoned", "failed", "in_progress"]:
                    raise Exception(
                        f"Workflow step {euid} is already {bobj.bstatus}, cannot set to {status}"
                    )

                if "step_properties" in bobj.json_addl:
                    bobj.json_addl["step_properties"]["start_operator"] = un
                    bobj.json_addl["step_properties"]["start_timestamp"] = now_dt
                bobj.json_addl["properties"]["status_timestamp"] = now_dt
                bobj.json_addl["properties"]["start_operator"] = un

            if status in ["complete", "abandoned", "failed"]:
                if bobj.bstatus in ["complete", "abandoned", "failed"]:
                    raise Exception(
                        f"Workflow step {euid} is already in a terminal {bobj.bstatus}, cannot set to {status}"
                    )

                bobj.json_addl["action_groups"][action_group]["actions"][action][
                    "action_enabled"
                ] = "0"

                if "step_properties" in bobj.json_addl:
                    bobj.json_addl["step_properties"]["end_operator"] = un
                    bobj.json_addl["step_properties"]["end_timestamp"] = now_dt

                bobj.json_addl["properties"]["end_timestamp"] = now_dt
                bobj.json_addl["properties"]["end_operator"] = un

            bobj.bstatus = status

            flag_modified(bobj, "json_addl")
            flag_modified(bobj, "bstatus")
            ##self.session.flush()
            self.session.commit()
        except Exception as e:
            self.logger.exception(f"ERROR: {e}")
            self.session.rollback()
            raise e

        return bobj

    def query_audit_log_by_euid(self, euid):
        return (
            self.session.query(self.Base.classes.audit_log)
            .filter(self.Base.classes.audit_log.rel_table_euid_fk == euid)
            .all()
        )

    def check_lineages_for_type(self, lineages, type_name, parent_or_child=None):
        if parent_or_child == "parent":
            for lin in lineages:
                if lin.parent_instance.type == type_name:
                    return True
        elif parent_or_child == "child":
            for lin in lineages:
                if lin.child_instance.type == type_name:
                    return True
        else:
            raise Exception("Must specify parent or child")

        return False

    # Backward compatibility alias
    check_lineages_for_btype = check_lineages_for_type

    def get_cost_of_euid_children(self, euid):
        tot_cost = 0
        ctr = 0
        for ec_tups in self.query_cost_of_all_children(euid):
            tot_cost += float(ec_tups[1])
            ctr += 1
        return tot_cost if ctr > 0 else "na"

        # Start with the provided EUID
        initial_instance = (
            self.session.query(self.Base.classes.generic_instance)
            .filter_by(euid=euid)
            .first()
        )
        if initial_instance:
            return traverse_and_calculate_children_cogs(initial_instance)
        else:
            return 0

    def get_cogs_to_produce_euid(self, euid):

        # Function to fetch and calculate the COGS for a given object
        def calculate_cogs(orm_instance):
            if (
                "cogs" not in orm_instance.json_addl
                or "state" not in orm_instance.json_addl["cogs"]
            ):
                raise ValueError(
                    f"COGS or state information missing for EUID: {orm_instance.euid}"
                )

            if orm_instance.json_addl["cogs"]["state"] != "active":
                return 0

            cost = float(orm_instance.json_addl["cogs"]["cost"])
            fractional_cost = float(
                orm_instance.json_addl["cogs"].get("fractional_cost", 1)
            )
            allocation_type = orm_instance.json_addl["cogs"].get(
                "allocation_type", "single"
            )

            active_children = len(
                [
                    child
                    for child in orm_instance.child_of_lineages
                    if "cogs" in child.json_addl
                    and child.json_addl["cogs"].get("state") == "active"
                ]
            )
            if active_children == 0:
                active_children = 1.0
            return cost * float(fractional_cost) / float(active_children)

        # Recursive function to traverse the graph and accumulate COGS
        def traverse_history_and_calculate_cogs(orm_instance):
            total_cogs = calculate_cogs(orm_instance)

            # Traverse child_of_lineages to find parent instances and accumulate their COGS
            for lineage in orm_instance.child_of_lineages:
                parent_instance = lineage.parent_instance
                if parent_instance:
                    total_cogs += traverse_history_and_calculate_cogs(parent_instance)

            return total_cogs

        # Start with the provided EUID
        initial_instance = (
            self.session.query(self.Base.classes.generic_instance)
            .filter_by(euid=euid)
            .first()
        )
        if initial_instance:
            return traverse_history_and_calculate_cogs(initial_instance)
        else:
            return 0


    def search_objs_by_addl_metadata(
        self,
        file_search_criteria,
        search_greedy=True,
        type=None,
        subtype=None,
        category=None,
    ):
        query = self.session.query(self.Base.classes.generic_instance)
        
        def create_datetime_filter(key, value, conditions):
            start_datetime = value.get('start')
            end_datetime = value.get('end', start_datetime)
            if start_datetime > end_datetime:
                self.logger.exception(f"ERROR: start_datetime {start_datetime} is greater than end_datetime {end_datetime}")
                raise Exception(f"ERROR: start_datetime {start_datetime} is greater than end_datetime {end_datetime}")
            
            if start_datetime and end_datetime:
                json_path = key.split("->")
                
                non_empty_condition = and_(
                    func.jsonb_extract_path_text(self.Base.classes.generic_instance.json_addl, *json_path) != '',
                    func.jsonb_extract_path_text(self.Base.classes.generic_instance.json_addl, *json_path).isnot(None)
                )
                
                datetime_condition = cast(
                    func.jsonb_extract_path_text(
                        self.Base.classes.generic_instance.json_addl, *json_path
                    ), 
                    DateTime
                ).between(start_datetime, end_datetime)
                
                combined_condition = and_(non_empty_condition, datetime_condition)
                
                conditions.append(combined_condition)
        
        def handle_jsonb_filter(key, value, conditions):
            if isinstance(value, dict):
                for sub_key, sub_value in value.items():
                    if sub_key.endswith('_datetime') and isinstance(sub_value, dict):
                        create_datetime_filter(f"{key}->{sub_key}", sub_value, conditions)
                    else:
                        if isinstance(sub_value, list):
                            for item in sub_value:
                                jsonb_filter = {key: {sub_key: item}}
                                conditions.append(
                                    self.Base.classes.generic_instance.json_addl.op("@>")(
                                        json.dumps(jsonb_filter, default=str)
                                    )
                                )
                        else:
                            jsonb_filter = {key: {sub_key: sub_value}}
                            conditions.append(
                                self.Base.classes.generic_instance.json_addl.op("@>")(
                                    json.dumps(jsonb_filter, default=str)
                                )
                            )
            else:
                if isinstance(value, list):
                    for item in value:
                        jsonb_filter = {key: item}
                        conditions.append(
                            self.Base.classes.generic_instance.json_addl.op("@>")(
                                json.dumps(jsonb_filter, default=str)
                            )
                        )
                else:
                    jsonb_filter = {key: value}
                    conditions.append(
                        self.Base.classes.generic_instance.json_addl.op("@>")(
                            json.dumps(jsonb_filter, default=str)
                        )
                    )

        if search_greedy:
            # Greedy search: matching any of the provided search keys
            or_conditions = []
            for key, value in file_search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )
                handle_jsonb_filter(key, value, or_conditions)
            if or_conditions:
                query = query.filter(or_(*or_conditions))
        else:
            # Non-greedy search: matching all specified search terms
            and_conditions = []
            for key, value in file_search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )
                handle_jsonb_filter(key, value, and_conditions)
            if and_conditions:
                query = query.filter(and_(*and_conditions))

        if type is not None:
            query = query.filter(self.Base.classes.generic_instance.type == type)

        if subtype is not None:
            query = query.filter(
                self.Base.classes.generic_instance.subtype == subtype
            )

        if category is not None:
            query = query.filter(
                self.Base.classes.generic_instance.category == category
            )

        logging.info(f"Generated SQL: {str(query.statement)}")

        results = query.all()
        return [result.euid for result in results]

    def search_objs_by_addl_metadataOG(
        self,
        file_search_criteria,
        search_greedy=True,
        type=None,
        subtype=None,
        category=None,
    ):
        query = self.session.query(self.Base.classes.generic_instance)
        
        def create_datetime_filter(key, value, conditions):
            start_datetime = value.get('start')
            end_datetime = value.get('end', start_datetime )
            if start_datetime > end_datetime:
                self.logger.exception(f"ERROR: start_datetime {start_datetime} is greater than end_datetime {end_datetime}")
                raise Exception(f"ERROR: start_datetime {start_datetime} is greater than end_datetime {end_datetime}")
            
            if start_datetime and end_datetime:
                json_path = key.split("->")
                
                non_empty_condition = and_(
                    func.jsonb_extract_path_text(self.Base.classes.generic_instance.json_addl, *json_path) != '',
                    func.jsonb_extract_path_text(self.Base.classes.generic_instance.json_addl, *json_path).isnot(None)
                )
                
                datetime_condition = cast(
                    func.jsonb_extract_path_text(
                        self.Base.classes.generic_instance.json_addl, *json_path
                    ), 
                    DateTime
                ).between(start_datetime, end_datetime)
                
                combined_condition = and_(non_empty_condition, datetime_condition)
                
                conditions.append(combined_condition)
            
        def handle_jsonb_filter(key, value, conditions):
            if isinstance(value, dict):
                for sub_key, sub_value in value.items():
                    if sub_key.endswith('_datetime') and isinstance(sub_value, dict):
                        create_datetime_filter(f"{key}->{sub_key}", sub_value, conditions)
                    else:
                        jsonb_filter = {key: {sub_key: sub_value}}
                        conditions.append(
                            self.Base.classes.generic_instance.json_addl.op("@>")(
                                json.dumps(jsonb_filter, default=str)
                            )
                        )
            else:
                jsonb_filter = {key: value}
                conditions.append(
                    self.Base.classes.generic_instance.json_addl.op("@>")(
                        json.dumps(jsonb_filter, default=str)
                    )
                )

        if search_greedy:
            # Greedy search: matching any of the provided search keys
            or_conditions = []
            for key, value in file_search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )
                handle_jsonb_filter(key, value, or_conditions)
            if or_conditions:
                query = query.filter(or_(*or_conditions))
        else:
            # Non-greedy search: matching all specified search terms
            and_conditions = []
            for key, value in file_search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )
                handle_jsonb_filter(key, value, and_conditions)
            if and_conditions:
                query = query.filter(and_(*and_conditions))

        if type is not None:
            query = query.filter(self.Base.classes.generic_instance.type == type)

        if subtype is not None:
            query = query.filter(
                self.Base.classes.generic_instance.subtype == subtype
            )

        if category is not None:
            query = query.filter(
                self.Base.classes.generic_instance.category == category
            )

        logging.info(f"Generated SQL: {str(query.statement)}")

        results = query.all()
        return [result.euid for result in results]



__all__ = [
    "BloomObj",
    "PGLOBAL",
    "logger",
]



"""Retired workflow helpers kept importable while queue-only Bloom settles."""

from __future__ import annotations

from typing import Any

from bloom_lims.domain.workflows import BloomWorkflow, BloomWorkflowStep


class BloomWorkflowMixin:
    """Retired workflow mixin placeholder."""


class BloomWorkflowStepMixin:
    """Retired workflow-step mixin placeholder."""


def _retired(*_args: Any, **_kwargs: Any) -> Any:
    raise RuntimeError("Bloom workflow runtime is retired; use queue-driven beta APIs instead.")


create_workflow = _retired
create_workflow_step = _retired
get_workflow_by_euid = _retired
get_workflow_steps = _retired
advance_workflow = _retired


__all__ = [
    "BloomWorkflow",
    "BloomWorkflowStep",
    "BloomWorkflowMixin",
    "BloomWorkflowStepMixin",
    "create_workflow",
    "create_workflow_step",
    "get_workflow_by_euid",
    "get_workflow_steps",
    "advance_workflow",
]



"""Schema coverage guard for action templates."""

from __future__ import annotations

import json
from pathlib import Path

ACTION_DIR = Path("bloom_lims/config/action")


def _iter_action_templates():
    for path in sorted(ACTION_DIR.glob("*.json")):
        payload = json.loads(path.read_text(encoding="utf-8"))
        for action_name, versions in payload.items():
            if not isinstance(versions, dict):
                continue
            for version, data in versions.items():
                if not isinstance(data, dict):
                    continue
                action_definition = data.get("action_definition")
                if isinstance(action_definition, dict):
                    yield path, action_name, version, action_definition


def test_every_action_template_has_ui_schema_fields():
    missing = []
    for path, action_name, version, action_definition in _iter_action_templates():
        ui_schema = action_definition.get("ui_schema")
        fields = ui_schema.get("fields") if isinstance(ui_schema, dict) else None
        if not isinstance(fields, list):
            missing.append(f"{path}:{action_name}:{version}: missing ui_schema.fields")
            continue

        for idx, field in enumerate(fields):
            if not isinstance(field, dict):
                missing.append(f"{path}:{action_name}:{version}: field[{idx}] is not an object")
                continue
            for key in ("name", "type", "required"):
                if key not in field:
                    missing.append(f"{path}:{action_name}:{version}: field[{idx}] missing '{key}'")

    assert not missing, "\n".join(missing)


def test_no_legacy_html_captured_data_fields_remain():
    offenders = []
    for path, action_name, version, action_definition in _iter_action_templates():
        captured_data = action_definition.get("captured_data")
        if not isinstance(captured_data, dict):
            continue

        for key, value in captured_data.items():
            if isinstance(key, str) and key.startswith("_"):
                offenders.append(f"{path}:{action_name}:{version}: legacy captured_data key '{key}'")
            if isinstance(value, str) and ("<input" in value or "<select" in value or "<textarea" in value):
                offenders.append(f"{path}:{action_name}:{version}: legacy HTML markup still present")

    assert not offenders, "\n".join(offenders)



"""
BLOOM LIMS Command Line Interface

Click-based CLI with tab completion support.

Usage:
    bloom --help              Show all available commands
    bloom db status           Show database status
    bloom gui                 Start the BLOOM web UI
    bloom config              Show current configuration
"""

import sys

import click
from rich.console import Console

from bloom_lims.cli.config_cmd import config
from bloom_lims.cli.db import db
from bloom_lims.cli.gui import gui, stop
from bloom_lims.cli.info import doctor, info, status, version
from bloom_lims.cli.integrations import integrations
from bloom_lims.cli.quality import quality
from bloom_lims.cli.server import server
from bloom_lims.cli.test import test
from bloom_lims.cli.users import users
from bloom_lims.cli.utils import logs, shell

console = Console()


def _get_version() -> str:
    """Get version from _version module."""
    try:
        from bloom_lims._version import get_version
        return get_version()
    except ImportError:
        return "dev"


@click.group(invoke_without_command=True)
@click.option('-v', '--verbose', is_flag=True, help='Enable verbose output')
@click.option('--version', 'show_version', is_flag=True, help='Show version and exit')
@click.pass_context
def cli(ctx, verbose, show_version):
    """BLOOM LIMS - Laboratory Information Management System CLI."""
    ctx.ensure_object(dict)
    ctx.obj['verbose'] = verbose

    if show_version:
        console.print(f"bloom [cyan]{_get_version()}[/cyan]")
        ctx.exit(0)

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


# Register command groups
cli.add_command(db)
cli.add_command(gui)
cli.add_command(stop)
cli.add_command(info)
cli.add_command(status)
cli.add_command(doctor)
cli.add_command(version)
cli.add_command(config)
cli.add_command(server)
cli.add_command(test)
cli.add_command(quality)
cli.add_command(users)
cli.add_command(integrations)
cli.add_command(shell)
cli.add_command(logs)


def main():
    """Main entry point for BLOOM CLI."""
    cli()


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from typing import Dict, List

from fastapi import APIRouter, Depends, Query, Request
from fastapi.responses import HTMLResponse

from bloom_lims.bobjs import BloomObj
from bloom_lims.db import BLOOMdb3
from bloom_lims.gui.deps import _is_tapdb_reachable, require_auth
from bloom_lims.gui.jinja import templates
from bloom_lims.search import SearchRequest, SearchService


router = APIRouter()


def _props(instance) -> dict:
    payload = instance.json_addl if isinstance(instance.json_addl, dict) else {}
    props = payload.get("properties", {})
    return props if isinstance(props, dict) else {}


def _is_queue_runtime_record(instance, beta_kind: str) -> bool:
    return str(_props(instance).get("beta_kind") or "").strip() == beta_kind


def _parse_csv_param(raw_value: str) -> List[str]:
    return [token.strip().lower() for token in (raw_value or "").split(",") if token.strip()]


def _render_search_page(
    *,
    request: Request,
    user_data: Dict,
    search_request: SearchRequest | None,
    query_value: str,
    categories_value: str,
    record_types_value: str,
) -> HTMLResponse:
    search_payload = {
        "query": query_value,
        "items": [],
        "total": 0,
        "page": 1,
        "page_size": 50,
        "total_pages": 1,
        "sort_by": "timestamp",
        "sort_order": "desc",
        "truncated": False,
        "facets": {"record_type": {}, "category": {}},
    }

    if search_request is not None:
        service = SearchService(username=user_data.get("email", "anonymous"))
        search_payload = service.search(search_request).model_dump(mode="json")

    template = templates.get_template("modern/search_results.html")
    context = {
        "request": request,
        "udat": user_data,
        "query": query_value,
        "types": categories_value,
        "record_types": record_types_value,
        "results": search_payload.get("items", []),
        "search_response": search_payload,
        "selected_types": _parse_csv_param(categories_value),
        "selected_record_types": _parse_csv_param(record_types_value),
    }
    return HTMLResponse(content=template.render(context), status_code=200)


@router.get("/", response_class=HTMLResponse)
async def modern_dashboard(request: Request, _=Depends(require_auth)):
    user_data = request.session.get("user_data", {})
    stats = {
        "queue_runtime_total": 0,
        "objects_total": 0,
        "equipment_total": 0,
        "reagents_total": 0,
    }
    recent_queue_runtime = []
    recent_objects = []
    db_unavailable = not _is_tapdb_reachable()

    if not db_unavailable:
        try:
            bobdb = BloomObj(BLOOMdb3(app_username=user_data.get("email", "anonymous")))
            generic_rows = (
                bobdb.session.query(bobdb.Base.classes.generic_instance)
                .filter_by(is_deleted=False)
                .all()
            )
            queue_definitions = [
                row for row in generic_rows if _is_queue_runtime_record(row, "queue_definition")
            ]
            open_work_items = [
                row
                for row in generic_rows
                if _is_queue_runtime_record(row, "beta_work_item")
                and str(getattr(row, "bstatus", "") or "").strip().lower() in {"open", "active"}
            ]
            stats = {
                "queue_runtime_total": len(open_work_items),
                "objects_total": len(generic_rows),
                "equipment_total": bobdb.session.query(bobdb.Base.classes.equipment_instance)
                .filter_by(is_deleted=False)
                .count(),
                "reagents_total": bobdb.session.query(bobdb.Base.classes.content_instance)
                .filter(
                    bobdb.Base.classes.content_instance.is_deleted == False,
                    bobdb.Base.classes.content_instance.subtype.like("%reagent%"),
                )
                .count(),
            }
            recent_queue_runtime = sorted(
                queue_definitions,
                key=lambda row: getattr(row, "created_dt", None),
                reverse=True,
            )[:5]
            recent_objects = (
                bobdb.session.query(bobdb.Base.classes.generic_instance)
                .filter_by(is_deleted=False)
                .order_by(bobdb.Base.classes.generic_instance.created_dt.desc())
                .limit(5)
                .all()
            )
        except Exception:
            db_unavailable = True

    template = templates.get_template("modern/dashboard.html")
    context = {
        "request": request,
        "udat": user_data,
        "stats": stats,
        "recent_queue_runtime": recent_queue_runtime,
        "recent_objects": recent_objects,
        "db_unavailable": db_unavailable,
    }

    return HTMLResponse(content=template.render(context), status_code=200)


@router.get("/create_object", response_class=HTMLResponse)
async def create_object_wizard(request: Request, _auth=Depends(require_auth)):
    user_data = request.session.get("user_data", {})
    template = templates.get_template("modern/create_object_wizard.html")
    context = {
        "request": request,
        "user_data": user_data,
        "user": user_data,
        "page_title": "Create Object",
    }
    return HTMLResponse(content=template.render(context), status_code=200)


@router.get("/help", response_class=HTMLResponse)
async def help_page(request: Request):
    user_data = request.session.get("user_data", {})
    template = templates.get_template("modern/help.html")
    context = {
        "request": request,
        "udat": user_data,
        "user": user_data,
    }
    return HTMLResponse(content=template.render(context), status_code=200)


@router.get("/search", response_class=HTMLResponse)
async def modern_search(
    request: Request,
    q: str = Query("", description="Search query"),
    types: str = Query("", description="Comma-separated categories to search"),
    record_types: str = Query("", description="Comma-separated record types to search"),
    page: int = Query(1, ge=1),
    page_size: int = Query(50, ge=1, le=200),
    sort_by: str = Query("timestamp"),
    sort_order: str = Query("desc"),
    _=Depends(require_auth),
):
    user_data = request.session.get("user_data", {})
    categories = _parse_csv_param(types)
    selected_record_types = _parse_csv_param(record_types)

    has_filters = bool(q.strip() or categories or selected_record_types)
    search_request = None
    if has_filters:
        search_request = SearchRequest(
            query=q,
            categories=categories,
            record_types=selected_record_types,
            page=page,
            page_size=page_size,
            sort_by=sort_by,
            sort_order=sort_order,
            max_scan=10000,
        )

    return _render_search_page(
        request=request,
        user_data=user_data,
        search_request=search_request,
        query_value=q,
        categories_value=types,
        record_types_value=record_types,
    )


@router.post("/search", response_class=HTMLResponse)
async def modern_search_from_dewey(request: Request, _=Depends(require_auth)):
    user_data = request.session.get("user_data", {})
    form = await request.form()
    form_data = {}
    for key in form.keys():
        values = form.getlist(key)
        form_data[key] = values if len(values) > 1 else values[0]

    target = str(form_data.get("search_target", "file")).strip().lower()
    service = SearchService(username=user_data.get("email", "anonymous"))
    search_request = service.build_dewey_request(form_data, target=target)

    query_value = search_request.query
    categories_value = ",".join(search_request.categories)
    record_types_value = ",".join(search_request.record_types)
    return _render_search_page(
        request=request,
        user_data=user_data,
        search_request=search_request,
        query_value=query_value,
        categories_value=categories_value,
        record_types_value=record_types_value,
    )


@router.get("/bulk_create_containers", response_class=HTMLResponse)
async def modern_bulk_create_containers(request: Request, _=Depends(require_auth)):
    user_data = request.session.get("user_data", {})
    template = templates.get_template("modern/bulk_create_containers.html")
    context = {"request": request, "udat": user_data}
    return HTMLResponse(content=template.render(context), status_code=200)



"""
BLOOM LIMS Domain - File Classes

This module contains file-related classes for file attachments,
data files, and file sets.

Extracted from bloom_lims/bobjs.py for better code organization.
"""

import os
import json
import logging
import subprocess
import socket
import hashlib
from datetime import UTC, datetime, timedelta
from pathlib import Path
import urllib.parse

import boto3
import requests
from botocore.exceptions import NoCredentialsError, PartialCredentialsError

from sqlalchemy import desc
from sqlalchemy.orm.attributes import flag_modified

from bloom_lims.domain.base import BloomObj
from bloom_lims.domain.utils import get_datetime_string, generate_random_string, _update_recursive

import re

logger = logging.getLogger(__name__)


class BloomFile(BloomObj):
    
    def __init__(self, bdb, bucket_prefix=None, is_deleted=False, cfg_printers=False, cfg_fedex=False):
        super().__init__(bdb,is_deleted=is_deleted, cfg_printers=cfg_printers, cfg_fedex=cfg_fedex)
    
        if bucket_prefix is None:
            bucket_prefix = os.environ.get(
                "BLOOM_DEWEY_S3_BUCKET_PREFIX", "set-a-bucket-prefix-in-the-dotenv-file"
            )

        self.bucket_prefix = bucket_prefix
        self._s3_client = None

    @property
    def s3_client(self):
        if self._s3_client is None:
            self._s3_client = boto3.client("s3")
        return self._s3_client

    def _euid_numeric_value(self, euid):
        euid_str = str(euid or "").strip().upper()
        if euid_str:
            # Meridian formats:
            #   CATEGORY-BODYCHECK
            #   SANDBOX:CATEGORY-BODYCHECK
            candidate = euid_str.split(":", 1)[-1]
            if "-" in candidate:
                _, body_check = candidate.split("-", 1)
                if len(body_check) >= 2:
                    body = body_check[:-1]  # drop checksum char
                    alphabet = "0123456789ABCDEFGHJKMNPQRSTVWXYZ"
                    if body and all(ch in alphabet for ch in body):
                        value = 0
                        for ch in body:
                            value = (value * 32) + alphabet.index(ch)
                        if value > 0:
                            return value

        # Legacy fallback for historical non-Meridian IDs like GX123.
        digits = re.sub("[^0-9]", "", euid_str)
        if digits:
            return int(digits)

        # Final deterministic fallback for malformed/unexpected IDs.
        digest = hashlib.sha1(euid_str.encode("utf-8")).digest()
        return int.from_bytes(digest[:8], byteorder="big", signed=False)

    def _derive_bucket_name(self, euid):
        euid_int = self._euid_numeric_value(euid)
        response = self.s3_client.list_buckets()
        buckets = response["Buckets"]
        matching_buckets = [
            bucket["Name"]
            for bucket in buckets
            if bucket["Name"].startswith(self.bucket_prefix)
        ]
        bucket_suffixes = sorted(
            [
                int(re.sub("[^0-9]", "", name.replace(self.bucket_prefix, "")))
                for name in matching_buckets
            ]
        )

        for i in range(len(bucket_suffixes) - 1):
            if bucket_suffixes[i] <= euid_int < bucket_suffixes[i + 1]:
                return f"{self.bucket_prefix}{bucket_suffixes[i]}"

        if euid_int >= bucket_suffixes[-1]:
            return f"{self.bucket_prefix}{bucket_suffixes[-1]}"

        raise Exception("No matching bucket found for the provided EUID.")

    def _determine_s3_key(self, euid, data_file_name):
        bucket_name = self._derive_bucket_name(euid)
        euid_numeric_part = self._euid_numeric_value(euid)
        response = self.s3_client.list_objects_v2(
            Bucket=bucket_name, Prefix="", Delimiter="/"
        )

        logging.debug(f"ListObjectsV2 Response: {response}")
        folders = sorted(
            [
                int(content["Prefix"].rstrip("/"))
                for content in response.get("CommonPrefixes", [])
            ]
        )

        if not folders:
            # If no folders are found, create a '0' folder
            self.s3_client.put_object(Bucket=bucket_name, Key="0/")
            folders = [0]

        for i in range(len(folders) - 1):
            if folders[i] <= euid_numeric_part < folders[i + 1]:
                folder_prefix = folders[i]
                break
        else:
            folder_prefix = folders[-1] if euid_numeric_part >= folders[-1] else 0

        logging.debug(f"Determined folder_prefix: {folder_prefix}")
        return f"{folder_prefix}/{euid}.{data_file_name.split('.')[-1]}"

    def DELME_check_s3_key_exists(self, bucket_name, s3_key):
        try:
            self.s3_client.head_object(Bucket=bucket_name, Key=s3_key)
            return True
        except self.s3_client.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise e

    def link_file_to_parent(self, child_euid, parent_euid):
        self.create_generic_instance_lineage_by_euids(child_euid, parent_euid)
        self.session.commit()
        
    def create_file(
        self,
        file_metadata={},
        file_data=None,
        file_name=None,
        url=None,
        full_path_to_file=None,
        s3_uri=None,
        create_locked=False,
        addl_tags={},
    ):
        """
        Create a file or import files from an S3 directory.

        :param file_metadata: Metadata to associate with the file(s).
        :param file_data: File data to upload (binary data).
        :param file_name: Name of the file.
        :param url: URL to fetch the file data.
        :param full_path_to_file: Local path to the file.
        :param s3_uri: S3 URI of the file or directory.
        :param create_locked: Whether to lock the file(s) after creation. Defaults to True.
        :param addl_tags: Additional tags for the file(s).
        :return: Created file object or list of file objects.
        """
        if s3_uri:
            # Detect if S3 URI is a directory
            s3_parsed_uri = re.match(r"s3://([^/]+)/(.+)", s3_uri)
            if not s3_parsed_uri:
                raise ValueError("Invalid s3_uri format. Expected format: s3://bucket_name/prefix")
            
            bucket_name, prefix = s3_parsed_uri.groups()

            try:
                response = self.s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix, Delimiter='/')
                files = response.get('Contents', [])
                
                # If more than one object or the URI ends with '/', treat it as a directory
                if len(files) > 1 or s3_uri.endswith('/'):
                    created_files = []
                    for file in files:
                        file_key = file['Key']
                        
                        # Skip directories
                        if file_key.endswith('/'):
                            continue
                        
                        file_name = file_key.split('/')[-1]
                        current_s3_uri = f"s3://{bucket_name}/{file_key}"
                        
                        # Create individual files for each item in the directory
                        created_file = self.create_file(
                            file_metadata=file_metadata,
                            s3_uri=current_s3_uri,
                            create_locked=create_locked,
                            addl_tags=addl_tags,
                        )
                        created_files.append(created_file)
                    
                    return created_files
                
                # Otherwise, process as a single file
                s3_uri = f"s3://{bucket_name}/{files[0]['Key']}" if files else s3_uri
            
            except Exception as e:
                raise Exception(f"Error detecting file or directory for S3 URI {s3_uri}: {e}")

        # Existing logic for processing a single file
        file_properties = {"properties": file_metadata}
        import_or_remote = file_metadata.get('import_or_remote', 'import')

        # Query for file template with proper error handling
        file_templates = self.query_template_by_component_v2("file", "file", "generic", "1.0")
        if not file_templates:
            raise Exception(
                "File template not found: file/file/generic/1.0. "
                "Please ensure the database is seeded with file templates (run: bloom init)."
            )

        new_file = self.create_instance(
            file_templates[0].euid,
            file_properties,
        )
        self.session.commit()

        new_file.json_addl["properties"]["current_s3_bucket_name"] = (
            self._derive_bucket_name(new_file.euid)
        )
        flag_modified(new_file, "json_addl")
        self.session.commit()

        if file_data or url or full_path_to_file or s3_uri:
            try:
                new_file = self.add_file_data(
                    new_file.euid,
                    file_data,
                    file_name,
                    url,
                    full_path_to_file,
                    s3_uri,
                    addl_tags=addl_tags,
                    import_or_remote=import_or_remote,
                )
            except Exception as e:
                logging.exception(f"Error adding file data: {e}")
                new_file.bstatus = "error adding file data"
                flag_modified(new_file, "json_addl")
                self.session.commit()
                raise Exception(e)
        else:
            logging.warning(f"No data provided for file creation or import skipped: {file_data, url}")
            new_file.bstatus = "awaiting file data"
            self.session.commit()

        if create_locked:
            self.lock_file(new_file.euid)

        return new_file


    def create_file_old(
        self,
        file_metadata={},
        file_data=None,
        file_name=None,
        url=None,
        full_path_to_file=None,
        s3_uri=None,
        create_locked=True,
        addl_tags={},
    ):
        file_properties = {"properties": file_metadata}

        import_or_remote = file_metadata['import_or_remote']

        new_file = self.create_instance(
            self.query_template_by_component_v2("file", "file", "generic", "1.0")[
                0
            ].euid,
            file_properties,
        )
        self.session.commit()

        # Special handling for patient_id
        if (
            "patient_id" in file_metadata
        ):
            patient_id = file_metadata["patient_id"]
            search_criteria = {"properties": {"patient_id": patient_id}}
            existing_euids = self.search_objs_by_addl_metadata(
                search_criteria,
                True,
                category="actor",
                type="generic",
                subtype="patient"
            )

            if existing_euids:
                # Create child relationships to existing objects
                for euid in existing_euids:
                    self.create_generic_instance_lineage_by_euids(euid, new_file.euid)
            else:
                # Create a new actor/generic/patient object
                new_patient = self.create_instance(
                    self.query_template_by_component_v2(
                        "actor", "generic", "patient", "1.0"
                    )[0].euid,
                    {"properties": {"patient_id": patient_id}},
                )
                self.session.commit()
                self.create_generic_instance_lineage_by_euids(
                    new_patient.euid, new_file.euid
                )

        new_file.json_addl["properties"]["current_s3_bucket_name"] = (
            self._derive_bucket_name(new_file.euid)
        )
        flag_modified(new_file, "json_addl")
        self.session.commit()

        if file_data or url or full_path_to_file or s3_uri:
            try:
                new_file = self.add_file_data(
                    new_file.euid, file_data, file_name, url, full_path_to_file, s3_uri, addl_tags=addl_tags, import_or_remote=import_or_remote
                )
            except Exception as e:
                logging.exception(f"Error adding file data: {e}")
                new_file.bstatus = "error adding file data"
                flag_modified(new_file, "json_addl")
                self.session.commit()
                raise Exception(e)
        else:
            logging.warning(f"No data provided for file creation, or import skipped ({import_or_remote}): {file_data, url}")
            new_file.bstatus = "awaiting file data"
            self.session.commit()

        if create_locked:
            self.lock_file(new_file.euid)

        return new_file


    def sanitize_tag(self, value, is_key=False):
        """
        Sanitize the tag key or value to conform to AWS tag requirements by replacing disallowed characters.

        Parameters:
        - value (str): The tag key or value to sanitize.
        - is_key (bool): If True, sanitize as a tag key (128-character limit). 
                        If False, sanitize as a tag value (256-character limit).

        Returns:
        - str: Sanitized tag key or value.
        """
        # AWS tag key or value allowed characters
        allowed_characters_regex = r'[^a-zA-Z0-9 _\.:/=+\-@]'
        
        # Replace disallowed characters with '_'
        sanitized_value = re.sub(allowed_characters_regex, '_', value)
        
        # Trim leading and trailing spaces (not allowed by AWS)
        sanitized_value = sanitized_value.strip()
        
        # Enforce maximum length
        max_length = 128 if is_key else 256
        sanitized_value = sanitized_value[:max_length]
        
        return sanitized_value

    def format_addl_tags(self, add_tags):
        if not isinstance(add_tags, dict):
            raise ValueError("Input must be a dictionary.")

        formatted_tags = []
        for key, value in add_tags.items():
            if not isinstance(value, str):
                raise ValueError(f"Value for key '{key}' must be a string.")
            formatted_tags.append(
                f"{self.sanitize_tag(key)}={self.sanitize_tag(value)}"
            )

        return "&".join(formatted_tags)
    
    def add_file_data(
        self,
        euid,
        file_data=None,
        file_name=None,
        url=None,
        full_path_to_file=None,
        s3_uri=None,
        addl_tags={},
        import_or_remote=None
    ):
        file_instance = self.get_by_euid(euid)
        s3_bucket_name = file_instance.json_addl["properties"]["current_s3_bucket_name"]
        file_properties = {}

        if import_or_remote in ["Remote", "remote"] and (file_data is not None or url is not None or full_path_to_file is not None):
            raise ValueError("Remote file management is only supported with internal S3 URI.")

        addl_tag_string = self.format_addl_tags(addl_tags)
        if len(addl_tag_string) > 0:
            addl_tag_string = f"&{addl_tag_string}"
        
        if file_name is None:
            if url:
                file_name = url.split("/")[-1]
            elif s3_uri:
                file_name = s3_uri.split("/")[-1]
            elif full_path_to_file:
                file_name = Path(full_path_to_file).name
            else:
                raise ValueError(
                    "file_name must be provided if file_data or url is passed without a filename."
                )

        file_suffix = file_name.split(".")[-1]
        s3_key = self._determine_s3_key(euid, file_name)

        # Check if a file with the same EUID already exists in the bucket
        s3_key_path = "/".join(s3_key.split("/")[:-1])
        s3_key_path = s3_key_path + "/" if len(s3_key_path) > 0 else ""


        existing_files = self.s3_client.list_objects_v2(
            Bucket=s3_bucket_name, Prefix=f"{s3_key_path}{euid}."
        )
        if "Contents" in existing_files:
            self.logger.exception(
                f"A file with PREFIX EUID {euid} already exists in bucket {s3_bucket_name} {s3_key_path}."
            )
            raise Exception(
                f"A file with EUID {euid} already exists in bucket {s3_bucket_name} {s3_key_path}."
            )

        if s3_uri:
            # Was just doing this for only remote s3 uris, but am going to leave them be for now
            # Check if a remote file with the same metadata already exists

            search_criteria = {"properties": {"current_s3_uri": s3_uri}}
            existing_euids = self.search_objs_by_addl_metadata(search_criteria, True, category="file", type="file", subtype="generic")
            
            if len(existing_euids) > 0:
                raise Exception(f"Remote file with URI {s3_uri} already exists in the database as {existing_euids}.")

            s3uri_bucket=s3_uri.split("/")[2]
            s3uri_key="/".join(s3_uri.split("/")[3:])
            # Store metadata for the remote file
            file_properties = {
                "remote_s3_uri": s3_uri,
                "original_file_name": file_name,
                "name": file_name,
                "original_file_size_bytes": None,  # Size is unknown for remote files
                "original_file_suffix": file_suffix,
                "original_file_data_type": "s3uri",
                "file_type": file_suffix,
                "current_s3_uri": s3_uri,
                "original_s3_uri": s3_uri,  
                "current_s3_key": s3uri_key,
                "current_s3_bucket_name": s3uri_bucket,
                "import_or_remote": 'remote',
            }
                        # Check if the object has the 'dewey_euid' tag
            try:
                existing_tags = self.s3_client.get_object_tagging(
                    Bucket=s3uri_bucket,
                    Key=s3uri_key
                )

                # Check if 'dewey_euid' is already present in the tags
                for tag in existing_tags.get("TagSet", []):
                    if tag["Key"] == "dewey_euid":
                        raise Exception(f"Object {s3_uri} already has a 'dewey_euid' tag with value {tag['Value']}.")
            except self.s3_client.exceptions.NoSuchKey:
                raise Exception(f"S3 object {s3_uri} does not exist.")
            except Exception as e:
                self.logger.exception(f"Error checking tags for S3 object {s3_uri}: {e}")
                raise Exception(f"Failed to check tags for S3 object: {e}")
            
                    
            # Construct the tags
            tagging = {
                'TagSet': [
                    {'Key': 'dewey_original_file_name', 'Value': self.sanitize_tag(file_name)},
                    {'Key': 'dewey_original_file_path', 'Value': 'N/A'},
                    {'Key': 'dewey_original_file_suffix', 'Value': self.sanitize_tag(file_suffix)},
                    {'Key': 'dewey_euid', 'Value': self.sanitize_tag(euid)},
                ]
            }
                    
            # Apply the tags to the existing object
            try:
                self.s3_client.put_object_tagging(
                    Bucket=s3uri_bucket,
                    Key=s3uri_key,
                    Tagging=tagging
                )
                self.logger.info(f"Tags successfully applied to S3 object {s3_uri}")
            except Exception as e:
                self.logger.exception(f"Error tagging existing S3 object {s3_uri}: {e}\n\n{tagging}")
                raise Exception(f"Failed to tag S3 object: {e}\n{tagging}")

            _update_recursive(file_instance.json_addl["properties"], file_properties)
            flag_modified(file_instance, "json_addl")
            self.session.commit()
            return file_instance

        try:
            if file_data:
                file_data.seek(0)  # Ensure the file pointer is at the beginning
                file_size = len(file_data.read())
                file_data.seek(0)  # Reset the file pointer after reading
                
                try:
                    self.s3_client.put_object(
                        Bucket=s3_bucket_name,
                        Key=s3_key,
                        Body=file_data,
                        Tagging=f"dewey_original_file_name={self.sanitize_tag(file_name)}&dewey_original_file_path=N/A&&dewey_original_file_suffix={self.sanitize_tag(file_suffix)}&dewey_euid={self.sanitize_tag(euid)}{addl_tag_string}"
                    )

                except Exception as e:
                    self.logger.exception(f"Error uploading file data: {e}. Possibly tag related: {self.sanitize_tag(file_name)}, {self.sanitize_tag(str(file_size))}, {self.sanitize_tag(file_suffix)}, {self.sanitize_tag(euid)} ") 
                    raise Exception(e)
                odirectory, ofilename = os.path.split(file_name)

                file_properties = {
                    "current_s3_key": s3_key,
                    "original_file_name": ofilename,
                    "name": file_name,
                    "original_file_path": odirectory,
                    "original_file_size_bytes": file_size,
                    "original_file_suffix": file_suffix,
                    "original_file_data_type": "raw data",
                    "file_type": file_suffix,
                    "current_s3_uri": f"s3://{s3_bucket_name}/{s3_key}",
                    "import_or_remote": import_or_remote,
                }

            elif url:
                response = requests.get(url)
                file_size = len(response.content)
                url_info = url.split("/")[-1]
                file_suffix = url_info.split(".")[-1]
                self.s3_client.put_object(
                    Bucket=s3_bucket_name,
                    Key=s3_key,
                    Body=response.content,
                    Tagging=f"dewey_original_file_name={self.sanitize_tag(url_info)}&dewey_original_url={self.sanitize_tag(url)}&dewey_original_file_suffix={self.sanitize_tag(file_suffix)}&dewey_euid={self.sanitize_tag(euid)}{addl_tag_string}",
                )
                file_properties = {
                    "current_s3_key": s3_key,
                    "original_file_name": url_info,
                    "name": url_info,
                    "original_url": url,
                    "original_file_size_bytes": file_size,
                    "original_file_suffix": file_suffix,
                    "original_file_data_type": "url",
                    "file_type": file_suffix,
                    "current_s3_uri": f"s3://{s3_bucket_name}/{s3_key}",
                    "import_or_remote": import_or_remote,
                }

            elif full_path_to_file:
                with open(full_path_to_file, "rb") as file:
                    file_data = file.read()
                file_size = os.path.getsize(full_path_to_file)
                local_path_info = Path(full_path_to_file)
                local_ip = None
                try:
                    local_ip = socket.gethostbyname(socket.gethostname())
                except socket.gaierror:
                    local_ip = "127.0.0.1"  # Fallback to localhost

                self.s3_client.put_object(
                    Bucket=s3_bucket_name,
                    Key=s3_key,
                    Body=file_data,
                    Tagging=f"dewey_original_file_name={self.sanitize_tag(local_path_info.name)}&dewey_original_file_path={self.sanitize_tag(full_path_to_file)}&dewey_original_file_suffix={self.sanitize_tag(file_suffix)}&dewey_euid={self.sanitize_tag(euid)}{addl_tag_string}",
                )
                file_properties = {
                    "current_s3_key": s3_key,
                    "original_file_name": local_path_info.name,
                    "name": local_path_info.name,
                    "original_file_path": full_path_to_file,
                    "original_local_server_name": socket.gethostname(),
                    "original_server_ip": local_ip,
                    "original_file_size_bytes": file_size,
                    "original_file_suffix": file_suffix,
                    "original_file_data_type": "local file",
                    "file_type": file_suffix,
                    "current_s3_uri": f"s3://{s3_bucket_name}/{s3_key}",
                    "import_or_remote": import_or_remote,
                }

            elif s3_uri in ["deprecate this"]:
                # I do not want to be in the business of moving files around here
                #elif s3_uri:
                # Validate and move the file from the provided s3_uri
                s3_parsed_uri = re.match(r"s3://([^/]+)/(.+)", s3_uri)
                if not s3_parsed_uri:
                    raise ValueError(
                        "Invalid s3_uri format. Expected format: s3://bucket_name/key"
                    )

                source_bucket, source_key = s3_parsed_uri.groups()
                try:
                    self.s3_client.head_object(Bucket=source_bucket, Key=source_key)
                except self.s3_client.exceptions.NoSuchKey:
                    raise ValueError(
                        f"The s3_uri {s3_uri} does not exist or is not accessible with the provided credentials."
                    )

                copy_source = {"Bucket": source_bucket, "Key": source_key}
                self.s3_client.copy(copy_source, s3_bucket_name, s3_key)
                file_size = self.s3_client.head_object(
                    Bucket=s3_bucket_name, Key=s3_key
                )["ContentLength"]

                file_properties = {
                    "current_s3_key": s3_key,
                    "original_file_name": file_name,
                    "name": file_name,
                    "original_s3_uri": s3_uri,
                    "original_file_size_bytes": file_size,
                    "original_file_suffix": file_suffix,
                    "original_file_data_type": "s3_uri",
                    "file_type": file_suffix,
                    "current_s3_uri": f"s3://{s3_bucket_name}/{s3_key}",
                    "import_or_remote": import_or_remote,
                }

                # Delete the old file and create a marker file
                marker_key = f"{source_key}.dewey.{euid}.moved"
                if len(marker_key)  >= 1024:
                    raise Exception(f"Marker key length is too long, >1024chrar : {len(marker_key)},not deleting original file: {source_key}")
                self.s3_client.put_object(
                    Bucket=source_bucket,
                    Key=marker_key,
                    Body=b"",
                    Tagging=f"dewey_import_or_remote={import_or_remote}&dewey_euid={euid}&dewey_original_s3_uri={self.sanitize_tag(s3_uri)}{addl_tag_string}",
                )
                #self.s3_client.delete_object(Bucket=source_bucket, Key=source_key)


            else:
                self.logger.exception("No file data provided.")
                raise ValueError("No file data provided.")

        except Exception as e:
            logging.exception(f"An error occurred while uploading the file: {e}")
            file_instance.bstatus = "error"
            file_instance.json_addl["properties"]["comments"] = (
                str(e) + f" FILENAM == {file_name}"
            )
            flag_modified(file_instance, "json_addl")
            flag_modified(file_instance, "bstatus")
            self.session.flush()
            self.session.commit()
            raise (e)

        _update_recursive(file_instance.json_addl["properties"], file_properties)
        flag_modified(file_instance, "json_addl")
        self.session.commit()

        return file_instance

    def update_file_metadata(self, euid, file_metadata={}):
        file_instance = self.get_by_euid(euid)
        _update_recursive(file_instance.json_addl["properties"], file_metadata)
        flag_modified(file_instance, "json_addl")
        self.session.commit()
        return file_instance

    def get_file_by_euid(self, euid):
        return self.get_by_euid(euid)

    def download_file(
        self,
        euid,
        save_pattern="dewey",
        include_metadata=False,
        save_path="./tmp/",
        delete_if_exists=False,
    ):
        """
        Downloads the S3 file locally with different naming patterns and optionally includes metadata in a YAML file.

        :param euid: EUID of the file to download.
        :param save_pattern: Naming pattern for the saved file. Options: 'dewey', 'orig', 'hybrid'.
        :param include_metadata: Whether to save metadata in a YAML file. Defaults to False.
        :param save_path: Directory where the file will be saved. Defaults to ./tmp/, which will be created if not present.
        :return: Path of the saved file.
        """
        import random
        random.randint(1,99999999)
        save_path = os.path.join(save_path, str(random.randint(1,99999999)))
        os.system(f"mkdir -p {save_path}")
        
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        else:
            self.logger.warn(f"Directory already exists: {save_path}")

        file_instance = self.get_by_euid(euid)
        s3_bucket_name = file_instance.json_addl["properties"]["current_s3_bucket_name"]
        s3_key = file_instance.json_addl["properties"]["current_s3_key"]
        original_file_name = file_instance.json_addl["properties"]["original_file_name"]
        file_suffix = file_instance.json_addl["properties"]["original_file_suffix"]

        if save_pattern == "dewey":
            local_file_name = f"{euid}.{file_suffix}"
        elif save_pattern == "orig":
            local_file_name = original_file_name
            print("WARNING: Using 'orig' pattern may overwrite existing files!")
        elif save_pattern == "hybrid":
            local_file_name = f"{euid}.{original_file_name}"
        else:
            raise ValueError(
                "Invalid save_pattern. Options are: 'dewey', 'orig', 'hybrid'."
            )

        local_file_path = os.path.join(save_path, local_file_name)

        if os.path.exists(local_file_path):
            self.logger.exception(f"File already exists: {local_file_path}")
            if delete_if_exists:
                os.remove(local_file_path)  # Delete the existing file
            else:
                raise Exception(f"File already exists: {local_file_path}")

        # Save metadata as a YAML file if requested
        if include_metadata:
            metadata_file_path = f"{local_file_path}.{euid}.dewey.yaml"
            if os.path.exists(metadata_file_path):
                self.logger.exception(
                    f"Metadata file already exists: {metadata_file_path}"
                )

                if delete_if_exists:
                    os.remove(metadata_file_path)
                else:
                    raise Exception(
                        f"Metadata file already exists: {metadata_file_path}"
                    )

            with open(metadata_file_path, "w") as metadata_file:
                yaml.dump(file_instance.json_addl["properties"], metadata_file)
            print(f"Metadata saved successfully: {metadata_file_path}")

        # Download the file from S3
        try:
            with open(local_file_path, "wb") as file:
                self.s3_client.download_fileobj(s3_bucket_name, s3_key, file)
            print(f"File downloaded successfully: {local_file_path}")
        except Exception as e:
            raise Exception(f"An error occurred while downloading the file: {e}")

        os.system(f"(sleep 2000 && rm -rf {save_path}) &")
        return local_file_path

    def get_s3_uris(self, euids, include_metadata=False):
        """
        Returns a dictionary of EUIDs to arrays containing their corresponding S3 URIs and optionally their metadata.

        :param euids: List of EUIDs to retrieve S3 URIs for.
        :param include_metadata: Boolean indicating whether to include metadata in the result.
        :return: Dictionary with EUID as key and array [S3 URI, metadata] as value.
        """
        euid_to_s3_data = {}

        for euid in euids:
            try:
                file_instance = self.get_by_euid(euid)
                s3_bucket_name = file_instance.json_addl["properties"][
                    "current_s3_bucket_name"
                ]
                s3_key = file_instance.json_addl["properties"]["current_s3_key"]
                s3_uri = f"s3://{s3_bucket_name}/{s3_key}"
                metadata = (
                    file_instance.json_addl["properties"] if include_metadata else None
                )
                euid_to_s3_data[euid] = [s3_uri, metadata]
            except Exception as e:
                self.logger.error(f"Error retrieving S3 URI for EUID {euid}: {e}")
                euid_to_s3_data[euid] = [None, None]  # or handle error as needed

        return euid_to_s3_data

    def delete_file(self, euid):
        # SOFT delete (S3 record is not deleted)

        file_instance = self.get_by_euid(euid)
        # s3_bucket_name = file_instance.json_addl['properties']['current_s3_bucket_name']
        # s3_key = file_instance.json_addl['properties']['current_s3_key']

        try:
            # self.s3_client.delete_object(Bucket=s3_bucket_name, Key=s3_key)
            self.delete_obj(file_instance)
            self.session.commit()
            return True
        except Exception as e:
            self.logger.error(f"Error deleting file {euid}: {e}")
            self.session.rollback()
            return False

    def get_s3_object_stream(self, euid):
        file_instance = self.get_file_by_euid(euid)
        s3_bucket_name = file_instance.json_addl["properties"]["current_s3_bucket_name"]
        s3_key = file_instance.json_addl["properties"]["current_s3_key"]

        try:
            response = self.s3_client.get_object(Bucket=s3_bucket_name, Key=s3_key)
            content_type = response["ContentType"]
            return response["Body"], content_type
        except self.s3_client.exceptions.NoSuchKey:
            raise Exception("File not found")
        except NoCredentialsError:
            raise Exception("Credentials not available")
        except Exception as e:
            raise Exception(e)

    def lock_file(self, euid, lock=True):
        """
        Locks or unlocks the specified S3 file.

        :param euid: EUID of the file to lock/unlock.
        :param lock: Boolean indicating whether to lock (True) or unlock (False) the file. Defaults to True.
        """
        file_instance = self.get_by_euid(euid)
        s3_bucket_name = file_instance.json_addl["properties"]["current_s3_bucket_name"]
        s3_key = file_instance.json_addl["properties"]["current_s3_key"]

        try:
            if lock:                
                self.s3_client.put_object_retention(
                    Bucket=s3_bucket_name,
                    Key=s3_key,
                    Retention={
                        'Mode': 'GOVERNANCE',
                        'RetainUntilDate': (datetime.now(UTC) + timedelta(days=36500)).isoformat()
                    }
                )

            return True
        except Exception as e:
            logging.exception(f"An error occurred while {'locking' if lock else 'unlocking'} the file: {e}")
            return False
        
    
    def create_presigned_url(self, 
                             file_euid,  
                             valid_duration=3600, 
                             comments="", status="active", 
                             file_set_euid=None):
        """
        Create a presigned url and create a shared reference to track this.

        :param file_euid: EUID of the file to create a shared reference for.
        :param valid_duration: Duration in seconds for which the reference is valid.
        :param comments: Additional comments for the reference.
        :param status: Status of the reference. Defaults to 'active'.
        :return: Created file reference instance.
        """
        
        # change to allow setting the start_datetime in the future.
        start_datetime = datetime.now(UTC)
        end_datetime = start_datetime + timedelta(seconds=valid_duration)
        
        file_instance = self.get_by_euid(file_euid)
        s3_bucket_name = file_instance.json_addl["properties"]["current_s3_bucket_name"]
        s3_key = file_instance.json_addl["properties"]["current_s3_key"]

        presigned_url = self.s3_client.generate_presigned_url(
            'get_object',
            Params={'Bucket': s3_bucket_name, 'Key': s3_key},
            ExpiresIn=valid_duration
        )

        file_ref_obj = BloomFileReference(self._bdb)
        file_reference = file_ref_obj.create_file_reference(
            file_euid=file_euid,
            reference_type="presigned",
            valid_duration=valid_duration,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
            comments=comments,
            status=status,
            presigned_url=presigned_url,
            file_set_euid=file_set_euid, 
            visibility=None
        )

        return {"file_reference": file_reference, "presigned_url": presigned_url}


    def import_files_from_s3_directory(self, s3_uri, file_metadata={}, create_locked=True):
        """
        Import all files from a specified S3 directory (not recursively).
        
        :param s3_uri: The S3 URI of the directory (e.g., s3://bucket_name/folder/).
        :param file_metadata: Metadata to associate with each imported file.
        :param create_locked: Whether to lock the files upon creation. Defaults to True.
        :return: List of created file objects.
        """
        # Parse S3 URI
        s3_parsed_uri = re.match(r"s3://([^/]+)/(.+)", s3_uri)
        if not s3_parsed_uri:
            raise ValueError("Invalid s3_uri format. Expected format: s3://bucket_name/folder/")
        
        bucket_name, prefix = s3_parsed_uri.groups()
        
        # List objects in the directory
        try:
            response = self.s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix, Delimiter='/')
            files = response.get('Contents', [])
        except Exception as e:
            raise Exception(f"Error listing S3 directory {s3_uri}: {e}")
        
        created_files = []
        for file in files:
            file_key = file['Key']
            
            # Skip directories
            if file_key.endswith('/'):
                continue
            
            file_name = file_key.split('/')[-1]
            current_s3_uri = f"s3://{bucket_name}/{file_key}"
            
            # Add metadata specific to the file
            individual_file_metadata = file_metadata.copy()
            individual_file_metadata['file_name'] = file_name
            
            try:
                # Create and add the file
                created_file = self.create_file(
                    file_metadata=individual_file_metadata,
                    s3_uri=current_s3_uri,
                    create_locked=create_locked
                )
                created_files.append(created_file)
            except Exception as e:
                logging.error(f"Error importing file {current_s3_uri}: {e}")
        
        return created_files




# As in expiring s3 links and so on. Potentially allow sharing of files with hosting protocols like SFTP, etc...

class BloomFileReference(BloomObj):
    
    def __init__(self, bdb, is_deleted=False, cfg_printers=False, cfg_fedex=False):
        super().__init__(bdb,is_deleted=is_deleted, cfg_printers=cfg_printers, cfg_fedex=cfg_fedex)


    def create_file_reference(self, file_euid=None, reference_type='presigned', visibility='public', valid_duration=0, start_datetime=None, end_datetime=None, comments="", status="active", presigned_url="", file_set_euid=None, rclone_config={}):
        """
        Create a shared file reference.

        :param file_euid: EUID of the file.
        :param reference_type: Type of reference. 'presigned' or 'rclone http'.
        :param visibility: 'public' or 'controlled'.
        :param valid_duration: Duration in seconds for which the reference is valid.
        :param start_datetime: (Optional) Start datetime for the reference. Defaults to now.
        :param end_datetime: (Optional) End datetime for the reference. Calculated from valid_duration if not provided.
        :param comments: Additional comments for the reference.
        :param status: Status of the reference. Defaults to 'active'.
        :param rclone_config: Configuration for rclone http serve {'port': 8080, 'host': '0.0.0.0', 'user': 'user', 'passwd': 'passwd', 'bucket':'xxx-dewey-0'}.
        :return: Created file reference instance.
        """
        
        start_datetime = start_datetime or datetime.now(UTC)
        end_datetime = end_datetime or (start_datetime + timedelta(seconds=valid_duration))
        
        file_reference_metadata = {
            "status": status,
            "comments": comments,
            "visibility": visibility,
            "reference_type": reference_type,
            "valid_duration": valid_duration,
            "start_datetime": start_datetime.isoformat(),
            "end_datetime": end_datetime.isoformat(),
            "presigned_url": presigned_url,
            "rclone_config": rclone_config
        }
        
        file_reference = self.create_instance(
            self.query_template_by_component_v2(
                "file", "shared_ref", "generic", "1.0"
            )[0].euid,
            {"properties": file_reference_metadata},
        )
        
        if reference_type.startswith('rclone'):
            # Start the rclone http serve

            filter_fn = f"logs/{file_reference.euid}_filter.txt"
            fh = open(filter_fn, "w")

            fs = self.get_by_euid(file_set_euid)
            for x in fs.parent_of_lineages:
                print(x.child_instance.euid)
                fh.write(f"+ {x.child_instance.json_addl['properties']['current_s3_key']}\n")
            fh.write('- *\n')
            fh.close()


            cmd = f"timeout {valid_duration} {reference_type} blms3:{rclone_config['bucket']} --filter-from logs/{file_reference.euid}_filter.txt --addr {rclone_config['host']}:{rclone_config['port']} --user {rclone_config['user']} --pass {rclone_config['passwd']} 2>&1 > logs/{file_reference.euid}_rclone.log &"
            logging.info(f"Starting rclone http serve with command: {cmd}")
            
            file_reference.json_addl['properties']['rclone_cmd'] = cmd
            flag_modified(file_reference, "json_addl")
            
            try:
                # Start the command in the background
                process = subprocess.Popen(
                    cmd,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )

                # Log that the process has started
                logging.info(f"rclone command started with PID: {process.pid}")

                # Optionally, you can wait a moment and then check if the process is still running
                process.communicate(timeout=5)
                if process.poll() is None:
                    logging.info(f"rclone is running successfully in the background.")
                    file_reference.json_addl['properties']['rclone_pid'] = process.pid
                    file_reference.json_addl['properties']['rclone_status'] = 'running'
                    flag_modified(file_reference, "json_addl")
                else:
                    file_reference.json_addl['properties']['rclone_status'] = 'error'
                    flag_modified(file_reference, "json_addl")
                    logging.error(f"rclone command failed to start properly. Error: {process.stderr.read().decode().strip()}")

            except subprocess.TimeoutExpired:
                logging.info(f"rclone command started and is running in the background.")
                file_reference.json_addl['properties']['rclone_pid'] = process.pid
                file_reference.json_addl['properties']['rclone_status'] = 'running bkgrnd'
                flag_modified(file_reference, "json_addl")
            except Exception as e:
                logging.error(f"An error occurred while starting rclone: {str(e)}")
                file_reference.json_addl['properties']['rclone_status'] = 'error'
                flag_modified(file_reference, "json_addl")

            self.session.commit()
            logging.info(f"{cmd} was executed... see logs")

            

            
        if file_euid not in [None]:
            self.create_generic_instance_lineage_by_euids(
                file_euid, file_reference.euid, reference_type
            )
            
        if file_set_euid not in [None]:
            self.create_generic_instance_lineage_by_euids(
                file_set_euid, file_reference.euid, "from_set"
            )



        self.session.commit()
        return file_reference
    

class BloomFileSet(BloomObj):
    def __init__(self, bdb, is_deleted=False, cfg_printers=False, cfg_fedex=False):
        super().__init__(bdb,is_deleted=is_deleted, cfg_printers=cfg_printers, cfg_fedex=cfg_fedex)

    def create_file_set(self, file_uids=[], file_set_metadata={}):
        file_set = self.create_instance(
            self.query_template_by_component_v2("file", "file_set", "generic", "1.0")[
                0
            ].euid,
            {"properties": file_set_metadata},
        )
        self.session.commit()

        return file_set

    def add_files_to_file_set(self, file_set_euid, file_euids=[]):
        file_set = self.get_by_euid(file_set_euid)
        for file_euid in file_euids:
            self.create_generic_instance_lineage_by_euids(file_set_euid, file_euid)
        self.session.commit()
        return file_set

    def get_file_set_by_euid(self, euid):
        return self.get_by_euid(euid)

    def remove_files_from_file_set(self, file_set_euid, file_euids=[]):
        file_set = self.get_by_euid(file_set_euid)

        # delete the lineage for each file to this file set
        for file_euid in file_euids:
            for i in file_set.child_of_lineages:
                if i.child_instance.euid == file_euid:
                    self.delete_obj(i)

        self.session.commit()
        return file_set

    def search_file_sets_by_metadata(self, search_criteria, greedy=True):
        """
        Search for file sets based on additional metadata.

        :param search_criteria: Dictionary containing the metadata to search for.
        :param greedy: Boolean indicating whether to perform a greedy search (matching any criteria) or not (matching all criteria).
        :return: List of EUIDs of matching file sets.
        """

        query = self.session.query(self.Base.classes.file_instance)

        if greedy:
            # Greedy search: matching any of the provided search keys
            or_conditions = []
            for key, value in search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )

                # Create conditions for JSONB key-value pairs
                if isinstance(value, dict):
                    for sub_key, sub_value in value.items():
                        jsonb_filter = {key: {sub_key: sub_value}}
                        or_conditions.append(
                            self.Base.classes.file_set_instance.json_addl.op("@>")(
                                jsonb_filter
                            )
                        )
                else:
                    jsonb_filter = {key: value}
                    or_conditions.append(
                        self.Base.classes.file_set_instance.json_addl.op("@>")(
                            jsonb_filter
                        )
                    )

            if or_conditions:
                query = query.filter(or_(*or_conditions))
        else:
            # Non-greedy search: matching all specified search terms
            and_conditions = []
            for key, value in search_criteria.items():
                if key == "file_metadata":
                    key = "properties"
                    logging.warning(
                        "The key 'file_metadata' is being treated as 'properties'."
                    )

                if isinstance(value, dict):
                    for sub_key, sub_value in value.items():
                        jsonb_filter = {key: {sub_key: sub_value}}
                        and_conditions.append(
                            self.Base.classes.file_set_instance.json_addl.op("@>")(
                                jsonb_filter
                            )
                        )
                else:
                    jsonb_filter = {key: value}
                    and_conditions.append(
                        self.Base.classes.file_set_instance.json_addl.op("@>")(
                            jsonb_filter
                        )
                    )

            if and_conditions:
                query = query.filter(and_(*and_conditions))

        logging.info(f"Generated SQL: {str(query.statement)}")

        results = query.all()
        return [result.euid for result in results]
 


__all__ = [
    "BloomFile",
    "BloomFileReference",
    "BloomFileSet",
]



"""
Tests for Analytics Dashboard Integration

These tests verify that analytics SQL views and dashboard definitions
are properly structured and contain required elements.
"""

import json
import os
import re
from pathlib import Path

import pytest


# Get the project root directory
PROJECT_ROOT = Path(__file__).parent.parent
ANALYTICS_DIR = PROJECT_ROOT / "analytics"


class TestAnalyticsDirectoryStructure:
    """Tests for analytics directory structure."""

    def test_analytics_directory_exists(self):
        """Test that analytics directory exists."""
        assert ANALYTICS_DIR.exists()
        assert ANALYTICS_DIR.is_dir()

    def test_sql_views_directory_exists(self):
        """Test that sql_views directory exists."""
        sql_dir = ANALYTICS_DIR / "sql_views"
        assert sql_dir.exists()

    def test_dashboards_directory_exists(self):
        """Test that dashboards directory exists."""
        dashboards_dir = ANALYTICS_DIR / "dashboards"
        assert dashboards_dir.exists()

    def test_metabase_directory_exists(self):
        """Test that metabase directory exists."""
        metabase_dir = ANALYTICS_DIR / "metabase"
        assert metabase_dir.exists()


class TestSQLViews:
    """Tests for SQL analytics views."""

    @pytest.fixture
    def sql_content(self):
        """Load install_views.sql content."""
        sql_path = ANALYTICS_DIR / "sql_views" / "install_views.sql"
        assert sql_path.exists(), "install_views.sql not found"
        return sql_path.read_text()

    def test_install_views_exists(self):
        """Test that install_views.sql exists."""
        sql_path = ANALYTICS_DIR / "sql_views" / "install_views.sql"
        assert sql_path.exists()

    def test_contains_sample_throughput_view(self, sql_content):
        """Test that v_sample_throughput view is defined."""
        assert "v_sample_throughput" in sql_content
        assert "CREATE" in sql_content

    def test_contains_workflow_bottlenecks_view(self, sql_content):
        """Test that v_workflow_bottlenecks view is defined."""
        assert "v_workflow_bottlenecks" in sql_content

    def test_contains_equipment_utilization_view(self, sql_content):
        """Test that v_equipment_utilization view is defined."""
        assert "v_equipment_utilization" in sql_content

    def test_contains_turnaround_times_view(self, sql_content):
        """Test that v_turnaround_times view is defined."""
        assert "v_turnaround_times" in sql_content

    def test_contains_audit_activity_view(self, sql_content):
        """Test that v_audit_activity view is defined."""
        assert "v_audit_activity" in sql_content

    def test_contains_object_counts_view(self, sql_content):
        """Test that v_object_counts view is defined."""
        assert "v_object_counts" in sql_content

    def test_contains_daily_summary_view(self, sql_content):
        """Test that v_daily_summary view is defined."""
        assert "v_daily_summary" in sql_content

    def test_views_use_generic_instance_table(self, sql_content):
        """Test that views reference generic_instance table."""
        assert "generic_instance" in sql_content

    def test_views_have_comments(self, sql_content):
        """Test that views have documentation comments."""
        assert "COMMENT ON VIEW" in sql_content


class TestDashboardDefinitions:
    """Tests for dashboard JSON definitions."""

    def test_operations_overview_exists(self):
        """Test that operations_overview.json exists."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "operations_overview.json"
        assert dashboard_path.exists()

    def test_turnaround_times_exists(self):
        """Test that turnaround_times.json exists."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "turnaround_times.json"
        assert dashboard_path.exists()

    def test_equipment_utilization_exists(self):
        """Test that equipment_utilization.json exists."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "equipment_utilization.json"
        assert dashboard_path.exists()

    def test_operations_overview_valid_json(self):
        """Test that operations_overview.json is valid JSON."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "operations_overview.json"
        content = dashboard_path.read_text()
        data = json.loads(content)  # Will raise if invalid
        assert "name" in data
        assert "cards" in data

    def test_dashboard_has_required_fields(self):
        """Test that dashboard definitions have required fields."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "operations_overview.json"
        data = json.loads(dashboard_path.read_text())
        
        assert "name" in data
        assert "description" in data
        assert "cards" in data
        assert isinstance(data["cards"], list)
        assert len(data["cards"]) > 0

    def test_dashboard_cards_have_required_fields(self):
        """Test that dashboard cards have required fields."""
        dashboard_path = ANALYTICS_DIR / "dashboards" / "operations_overview.json"
        data = json.loads(dashboard_path.read_text())
        
        for card in data["cards"]:
            assert "id" in card
            assert "title" in card
            assert "type" in card
            assert "query" in card


class TestMetabaseSetup:
    """Tests for Metabase docker-compose setup."""

    def test_docker_compose_exists(self):
        """Test that docker-compose.yml exists."""
        compose_path = ANALYTICS_DIR / "metabase" / "docker-compose.yml"
        assert compose_path.exists()

    def test_run_script_exists(self):
        """Test that run_metabase.sh exists."""
        script_path = ANALYTICS_DIR / "metabase" / "run_metabase.sh"
        assert script_path.exists()

    def test_run_script_is_executable(self):
        """Test that run_metabase.sh is executable."""
        script_path = ANALYTICS_DIR / "metabase" / "run_metabase.sh"
        assert os.access(script_path, os.X_OK)

    def test_docker_compose_defines_metabase_service(self):
        """Test that docker-compose defines metabase service."""
        compose_path = ANALYTICS_DIR / "metabase" / "docker-compose.yml"
        content = compose_path.read_text()
        assert "metabase" in content.lower()
        assert "services" in content




"""
BLOOM LIMS Test Configuration and Fixtures

This module provides pytest fixtures and utilities for testing BLOOM LIMS.

Usage:
    # In test files, fixtures are automatically available
    def test_something(bdb, bloom_obj):
        result = bloom_obj.get_by_euid("WF_ABC123_X")
        assert result is not None

Fixtures:
    - bdb: Database connection (session-scoped)
    - bloom_obj: BloomObj instance
    - test_template: A test template for creating instances
    - clean_test_data: Cleanup fixture for test data
"""

import os
import secrets
import tempfile
from pathlib import Path

import pytest
import logging
from typing import Generator, Optional
from unittest.mock import MagicMock, patch

# Configure test logging
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


# Environment setup for tests
def pytest_configure(config):
    """Configure pytest environment."""
    # Enforce TapDB strict namespace mode for tests (v2 config).
    os.environ.setdefault("TAPDB_CLIENT_ID", "bloom")
    os.environ.setdefault("TAPDB_DATABASE_NAME", "bloom")
    os.environ.setdefault("TAPDB_STRICT_NAMESPACE", "1")

    # Provide a deterministic v2 namespaced TapDB config for CI/local runs
    # so tests do not depend on developer home config.
    if not (os.environ.get("TAPDB_CONFIG_PATH") or "").strip():
        local_port = str(os.environ.get("BLOOM_TAPDB_LOCAL_PG_PORT") or "5566").strip()
        user = str(os.environ.get("USER") or "postgres").strip()
        tmp_path = Path(tempfile.gettempdir()) / f"bloom_tapdb_config_{secrets.token_hex(16)}.yaml"
        tmp_path.write_text(
            "\n".join(
                [
                    "meta:",
                    "  config_version: 2",
                    "  client_id: bloom",
                    "  database_name: bloom",
                    "environments:",
                    "  dev:",
                    "    engine_type: local",
                    "    host: localhost",
                    f"    port: \"{local_port}\"",
                    "    ui_port: \"8912\"",
                    f"    user: \"{user}\"",
                    "    password: \"\"",
                    "    database: \"tapdb_bloom_dev\"",
                    "    cognito_user_pool_id: \"\"",
                    "    audit_log_euid_prefix: \"TAG\"",
                    "    support_email: \"support@dyly.bio\"",
                    "  test:",
                    "    engine_type: local",
                    "    host: localhost",
                    f"    port: \"{local_port}\"",
                    "    ui_port: \"8912\"",
                    f"    user: \"{user}\"",
                    "    password: \"\"",
                    "    database: \"tapdb_bloom_test\"",
                    "    cognito_user_pool_id: \"\"",
                    "    audit_log_euid_prefix: \"TAG\"",
                    "    support_email: \"support@dyly.bio\"",
                    "",
                ]
            ),
            encoding="utf-8",
        )
        os.chmod(tmp_path, 0o600)
        os.environ["TAPDB_CONFIG_PATH"] = str(tmp_path)

    # Set test database port if not already set
    if "PGPORT" not in os.environ:
        try:
            from bloom_lims.config import get_tapdb_db_config

            os.environ["PGPORT"] = str(get_tapdb_db_config().get("port") or "5566")
        except Exception:
            os.environ["PGPORT"] = "5566"

    # Disable SQL echo during tests unless explicitly enabled
    if "ECHO_SQL" not in os.environ:
        os.environ["ECHO_SQL"] = "False"

    # Disable rate limiting during tests to prevent 429 errors
    os.environ["BLOOM_DISABLE_RATE_LIMITING"] = "1"


@pytest.fixture(scope="session")
def bdb():
    """
    Session-scoped database connection fixture.
    
    Creates a single database connection for all tests in the session.
    The connection is automatically closed after all tests complete.
    
    Yields:
        BLOOMdb3: Database connection instance
    """
    from bloom_lims.db import BLOOMdb3
    
    logger.info("Creating test database connection")
    db = BLOOMdb3()
    
    yield db
    
    logger.info("Closing test database connection")
    db.close()


@pytest.fixture(scope="function")
def bdb_function():
    """
    Function-scoped database connection fixture.
    
    Creates a new database connection for each test function.
    Use this when tests need isolated database state.
    
    Yields:
        BLOOMdb3: Database connection instance
    """
    from bloom_lims.db import BLOOMdb3
    
    db = BLOOMdb3()
    yield db
    db.close()


@pytest.fixture(scope="function")
def bloom_obj(bdb):
    """
    BloomObj instance fixture.
    
    Creates a BloomObj instance using the session-scoped database connection.
    
    Args:
        bdb: Database connection fixture
        
    Yields:
        BloomObj: BloomObj instance
    """
    from bloom_lims.bobjs import BloomObj
    
    return BloomObj(bdb)


@pytest.fixture(scope="function")
def bloom_container(bdb):
    """BloomContainer instance fixture."""
    from bloom_lims.bobjs import BloomContainer
    return BloomContainer(bdb)


@pytest.fixture(scope="function")
def bloom_workflow(bdb):
    """BloomWorkflow instance fixture."""
    from bloom_lims.bobjs import BloomWorkflow
    return BloomWorkflow(bdb)


@pytest.fixture(scope="function")
def bloom_content(bdb):
    """BloomContent instance fixture."""
    from bloom_lims.bobjs import BloomContent
    return BloomContent(bdb)


@pytest.fixture(scope="function")
def test_template(bdb, bloom_obj):
    """
    Get a test template for creating instances.
    
    Returns the first available generic_template.
    
    Args:
        bdb: Database connection fixture
        bloom_obj: BloomObj fixture
        
    Returns:
        Template object or None if no templates exist
    """
    templates = bdb.session.query(bloom_obj.Base.classes.generic_template).limit(1).all()
    return templates[0] if templates else None


@pytest.fixture(scope="function")
def mock_session():
    """
    Mock SQLAlchemy session for unit tests.
    
    Use this when you want to test without database access.
    
    Returns:
        MagicMock: Mocked session object
    """
    session = MagicMock()
    session.query.return_value.filter.return_value.first.return_value = None
    session.query.return_value.filter_by.return_value.first.return_value = None
    return session


@pytest.fixture(scope="function")
def clean_cache():
    """
    Clear the global cache before and after test.
    
    Use this when testing cache behavior.
    """
    from bloom_lims.core.cache import cache_clear
    
    cache_clear()
    yield
    cache_clear()


# API Test Fixtures
@pytest.fixture(scope="function")
def api_client():
    """
    FastAPI test client fixture.

    Creates a test client for the BLOOM LIMS API.

    Yields:
        TestClient: FastAPI test client
    """
    from fastapi.testclient import TestClient
    from bloom_lims.main import app

    return TestClient(app)


@pytest.fixture(scope="function")
def authenticated_api_client(api_client):
    """
    Authenticated API client fixture.

    Creates a test client with mock authentication headers.

    Args:
        api_client: Base API client fixture

    Yields:
        TestClient: Authenticated FastAPI test client
    """
    # Add mock auth header
    api_client.headers["X-API-Key"] = "test-api-key"
    return api_client


@pytest.fixture(scope="function")
def mock_api_auth():
    """
    Mock API authentication for testing.

    Patches the require_api_auth dependency to return a test user.

    Yields:
        APIUser: Mock user object
    """
    from bloom_lims.api.v1.dependencies import APIUser

    mock_user = APIUser(
        user_id="test-user-id",
        email="test@example.com",
        role="admin",
    )

    with patch("bloom_lims.api.v1.dependencies.require_api_auth", return_value=mock_user):
        yield mock_user


# Test markers
def pytest_collection_modifyitems(config, items):
    """Add markers to tests based on their location."""
    for item in items:
        # Mark integration tests
        if "integration" in str(item.fspath):
            item.add_marker(pytest.mark.integration)

        # Mark slow tests
        if "slow" in item.name.lower():
            item.add_marker(pytest.mark.slow)

        # Mark API tests
        if "api" in str(item.fspath) or "api" in item.name.lower():
            item.add_marker(pytest.mark.api)



"""
Supplemental template seeding for Bloom legacy template configs.

TapDB's built-in seed command only loads its core/workflow template set. Bloom
still ships additional template definitions in `bloom_lims/config` using the
legacy file format. This module loads those templates and upserts them into
`generic_template` through the TapDB-backed Bloom adapter.
"""

from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterator, Optional

from daylily_tapdb.sequences import ensure_instance_prefix_sequence

from bloom_lims.db import BLOOMdb3


KNOWN_POLYMORPHIC_CATEGORIES = {
    "generic",
    "workflow",
    "workflow_step",
    "container",
    "content",
    "equipment",
    "data",
    "test_requisition",
    "actor",
    "action",
    "health_event",
    "file",
    "subject",
}


@dataclass(frozen=True)
class TemplateRecord:
    category: str
    type_name: str
    subtype: str
    version: str
    name: str
    polymorphic_discriminator: str
    instance_prefix: str
    instance_polymorphic_identity: str
    bstatus: str
    is_singleton: bool
    json_addl: dict


@dataclass(frozen=True)
class SeedSummary:
    templates_loaded: int
    inserted: int
    updated: int
    prefixes_ensured: int


def _default_config_dir() -> Path:
    return Path(__file__).resolve().parents[1] / "config"


def _coerce_bool(value: object) -> bool:
    if isinstance(value, bool):
        return value
    if value is None:
        return False
    normalized = str(value).strip().lower()
    return normalized in {"1", "true", "yes", "y", "on"}


def _normalize_prefix(prefix: object) -> str:
    if prefix is None:
        return ""
    normalized = str(prefix).strip().upper()
    return normalized if normalized.isalpha() else ""


def _template_polymorphic_discriminator(category: str) -> str:
    if category in KNOWN_POLYMORPHIC_CATEGORIES:
        return f"{category}_template"
    return "generic_template"


def _instance_polymorphic_identity(category: str) -> str:
    if category in KNOWN_POLYMORPHIC_CATEGORIES:
        return f"{category}_instance"
    return "generic_instance"


def _load_metadata_prefixes(category_dir: Path) -> Dict[str, str]:
    metadata_path = category_dir / "metadata.json"
    if not metadata_path.exists():
        return {}
    try:
        with open(metadata_path, "r", encoding="utf-8") as handle:
            payload = json.load(handle)
    except Exception:
        return {}

    euid_prefixes = payload.get("euid_prefixes", {})
    if not isinstance(euid_prefixes, dict):
        return {}

    result: Dict[str, str] = {}
    for key, value in euid_prefixes.items():
        normalized = _normalize_prefix(value)
        if normalized:
            result[str(key)] = normalized
    return result


def _resolve_instance_prefix(
    metadata_prefixes: Dict[str, str],
    type_name: str,
    subtype: str,
    template_data: dict,
) -> str:
    direct = _normalize_prefix(template_data.get("instance_prefix"))
    if direct:
        return direct

    for key in (type_name, subtype, "default"):
        if key in metadata_prefixes:
            value = _normalize_prefix(metadata_prefixes.get(key))
            if value:
                return value

    return "GX"


def _iter_legacy_template_records(config_dir: Path) -> Iterator[TemplateRecord]:
    for category_dir in sorted(config_dir.iterdir()):
        if not category_dir.is_dir() or category_dir.name.startswith("_"):
            continue

        category = category_dir.name
        metadata_prefixes = _load_metadata_prefixes(category_dir)

        for json_file in sorted(category_dir.glob("*.json")):
            if json_file.name == "metadata.json":
                continue

            type_name = json_file.stem
            with open(json_file, "r", encoding="utf-8") as handle:
                payload = json.load(handle)

            if not isinstance(payload, dict):
                continue

            for subtype, versions in payload.items():
                if not isinstance(versions, dict):
                    continue
                for version, template_data in versions.items():
                    if not isinstance(template_data, dict):
                        continue

                    code = f"{category}/{type_name}/{subtype}/{version}"
                    description = str(template_data.get("description") or code)
                    instance_prefix = _resolve_instance_prefix(
                        metadata_prefixes=metadata_prefixes,
                        type_name=type_name,
                        subtype=str(subtype),
                        template_data=template_data,
                    )

                    yield TemplateRecord(
                        category=category,
                        type_name=type_name,
                        subtype=str(subtype),
                        version=str(version),
                        name=description,
                        polymorphic_discriminator=_template_polymorphic_discriminator(category),
                        instance_prefix=instance_prefix,
                        instance_polymorphic_identity=_instance_polymorphic_identity(category),
                        bstatus=str(template_data.get("bstatus") or "active"),
                        is_singleton=_coerce_bool(
                            template_data.get("is_singleton", template_data.get("singleton"))
                        ),
                        json_addl=template_data,
                    )


def _ensure_instance_prefix_sequence(session, prefix: str) -> None:
    normalized = _normalize_prefix(prefix)
    if not normalized:
        return

    ensure_instance_prefix_sequence(session, normalized)


def seed_bloom_templates(config_dir: Optional[Path] = None) -> SeedSummary:
    """
    Upsert Bloom legacy templates into TAPDB generic_template.

    Returns summary counts for inserted/updated templates and ensured prefixes.
    """
    cfg_dir = Path(config_dir) if config_dir else _default_config_dir()
    if not cfg_dir.exists():
        raise FileNotFoundError(f"Bloom template config directory not found: {cfg_dir}")

    records = list(_iter_legacy_template_records(cfg_dir))
    if not records:
        return SeedSummary(templates_loaded=0, inserted=0, updated=0, prefixes_ensured=0)

    bdb = BLOOMdb3(app_username="bloom-template-seed")
    inserted = 0
    updated = 0
    prefixes: set[str] = set()

    try:
        session = bdb.session
        classes = bdb.Base.classes
        generic_template_cls = classes.generic_template

        for record in records:
            prefixes.add(record.instance_prefix)
            existing = (
                session.query(generic_template_cls)
                .filter(
                    generic_template_cls.category == record.category,
                    generic_template_cls.type == record.type_name,
                    generic_template_cls.subtype == record.subtype,
                    generic_template_cls.version == record.version,
                )
                .one_or_none()
            )

            if existing is None:
                model_cls = getattr(
                    classes, f"{record.category}_template", generic_template_cls
                )
                new_template = model_cls(
                    name=record.name,
                    polymorphic_discriminator=record.polymorphic_discriminator,
                    category=record.category,
                    type=record.type_name,
                    subtype=record.subtype,
                    version=record.version,
                    instance_prefix=record.instance_prefix,
                    instance_polymorphic_identity=record.instance_polymorphic_identity,
                    bstatus=record.bstatus,
                    is_singleton=record.is_singleton,
                    is_deleted=False,
                    json_addl=record.json_addl,
                )
                session.add(new_template)
                inserted += 1
                continue

            # Existing templates are intentionally left unchanged here. Some
            # local dev databases still carry legacy column typing, and
            # update-flushes through ORM mappings can fail in mixed-schema
            # environments. The supplemental seed path is used to ensure
            # required templates exist, not to mutate existing definitions.
            continue

        for prefix in sorted(prefixes):
            _ensure_instance_prefix_sequence(session, prefix)

        session.commit()
    except Exception:
        bdb.session.rollback()
        raise
    finally:
        bdb.close()

    return SeedSummary(
        templates_loaded=len(records),
        inserted=inserted,
        updated=updated,
        prefixes_ensured=len(prefixes),
    )



"""Deprecated container helpers rewritten to use graph edges, not JSON links."""

from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Any

from sqlalchemy import select
from sqlalchemy.orm import Session

from bloom_lims.exceptions import DatabaseError, NotFoundError, ValidationError

logger = logging.getLogger(__name__)


@dataclass
class ContainerPosition:
    """Represents a position within a container."""

    row: int
    column: int

    @classmethod
    def from_well(cls, well: str) -> "ContainerPosition":
        well = well.upper().strip()
        match = re.match(r"^([A-Z])(\\d+)$", well)
        if not match:
            raise ValidationError(f"Invalid well format: {well}", field="position")
        row = ord(match.group(1)) - ord("A")
        column = int(match.group(2)) - 1
        return cls(row=row, column=column)

    @classmethod
    def from_index(cls, index: int, num_columns: int) -> "ContainerPosition":
        return cls(row=index // num_columns, column=index % num_columns)

    def to_well(self) -> str:
        return f"{chr(ord('A') + self.row)}{self.column + 1}"

    def to_index(self, num_columns: int) -> int:
        return self.row * num_columns + self.column

    def __str__(self) -> str:
        return self.to_well()


def get_container_layout(container: Any) -> dict[str, Any]:
    if not container or not hasattr(container, "json_addl"):
        return {"rows": 1, "columns": 1, "type": "unknown"}

    payload = container.json_addl or {}
    layout = payload.get("layout", {}) if isinstance(payload, dict) else {}
    rows = int(layout.get("rows", 8) or 8)
    columns = int(layout.get("columns", 12) or 12)
    return {
        "rows": rows,
        "columns": columns,
        "type": container.subtype or container.type,
        "total_positions": rows * columns,
    }


def get_container_contents(
    session: Session,
    base,
    container_euid: str,
) -> list[dict[str, Any]]:
    logger.debug("Getting contents of container: %s", container_euid)
    container = _get_instance(session, base, container_euid)
    if container is None:
        raise NotFoundError(
            f"Container not found: {container_euid}",
            resource_type="container",
            resource_id=container_euid,
        )

    stmt = (
        select(base.classes.generic_instance_lineage)
        .where(
            base.classes.generic_instance_lineage.parent_instance_uid == container.uid,
            base.classes.generic_instance_lineage.relationship_type == "contains",
            base.classes.generic_instance_lineage.is_deleted.is_(False),
        )
        .order_by(base.classes.generic_instance_lineage.created_dt.asc())
    )
    rows = []
    for lineage in session.execute(stmt).scalars().all():
        child = getattr(lineage, "child_instance", None)
        if child is None or getattr(child, "is_deleted", False):
            continue
        payload = getattr(lineage, "json_addl", {}) or {}
        props = payload.get("properties") if isinstance(payload, dict) else {}
        rows.append(
            {
                "position": str((props or {}).get("position") or "").strip() or None,
                "euid": child.euid,
                "name": child.name,
                "type": child.type,
            }
        )
    return rows


def place_in_container(
    session: Session,
    base,
    container_euid: str,
    object_euid: str,
    position: str,
) -> bool:
    logger.debug("Placing %s in %s at %s", object_euid, container_euid, position)
    try:
        container = _get_instance(session, base, container_euid)
        if container is None:
            raise NotFoundError(
                f"Container not found: {container_euid}",
                resource_type="container",
                resource_id=container_euid,
            )
        obj = _get_instance(session, base, object_euid)
        if obj is None:
            raise NotFoundError(
                f"Object not found: {object_euid}",
                resource_type="object",
                resource_id=object_euid,
            )

        normalized_position = str(position or "").strip().upper()
        ContainerPosition.from_well(normalized_position)

        for entry in get_container_contents(session, base, container_euid):
            if entry.get("position") == normalized_position:
                raise ValidationError(
                    f"Position {normalized_position} is already occupied",
                    field="position",
                )

        lineage = base.classes.generic_instance_lineage(
            name=f"{container.name} :: {obj.name}",
            parent_type=(
                f"{container.category}:{container.type}:{container.subtype}:{container.version}"
            ),
            child_type=f"{obj.category}:{obj.type}:{obj.subtype}:{obj.version}",
            relationship_type="contains",
            parent_instance_uid=container.uid,
            child_instance_uid=obj.uid,
            category="generic",
            type="container",
            subtype="contains",
            version="1.0",
            polymorphic_discriminator="generic_instance_lineage",
            bstatus="active",
            json_addl={
                "properties": {
                    "position": normalized_position,
                    "placed_at": datetime.now(UTC).isoformat(),
                }
            },
        )
        session.add(lineage)
        session.flush()
        return True
    except (NotFoundError, ValidationError):
        raise
    except Exception as exc:  # pragma: no cover - defensive
        logger.error("Error placing object in container: %s", exc)
        raise DatabaseError(f"Failed to place object in container: {exc}")


def remove_from_container(
    session: Session,
    base,
    container_euid: str,
    position: str,
) -> str | None:
    logger.debug("Removing object from %s at %s", container_euid, position)
    container = _get_instance(session, base, container_euid)
    if container is None:
        raise NotFoundError(
            f"Container not found: {container_euid}",
            resource_type="container",
            resource_id=container_euid,
        )

    normalized_position = str(position or "").strip().upper()
    stmt = (
        select(base.classes.generic_instance_lineage)
        .where(
            base.classes.generic_instance_lineage.parent_instance_uid == container.uid,
            base.classes.generic_instance_lineage.relationship_type == "contains",
            base.classes.generic_instance_lineage.is_deleted.is_(False),
        )
        .order_by(base.classes.generic_instance_lineage.created_dt.asc())
    )
    for lineage in session.execute(stmt).scalars().all():
        payload = getattr(lineage, "json_addl", {}) or {}
        props = payload.get("properties") if isinstance(payload, dict) else {}
        if str((props or {}).get("position") or "").strip().upper() != normalized_position:
            continue
        lineage.is_deleted = True
        session.flush()
        child = getattr(lineage, "child_instance", None)
        return getattr(child, "euid", None)
    return None


def _get_instance(session: Session, base, euid: str):
    return (
        session.query(base.classes.generic_instance)
        .filter(
            base.classes.generic_instance.euid == str(euid or "").strip().upper(),
            base.classes.generic_instance.is_deleted.is_(False),
        )
        .first()
    )


try:  # pragma: no cover - compatibility export only
    from bloom_lims.bobjs import BloomObj as _BloomObj

    BloomContainer = _BloomObj
except ImportError:  # pragma: no cover - defensive
    BloomContainer = None
